"""Unit tests for the DraftDiffCommitResource."""

from django.core.files.uploadedfile import SimpleUploadedFile
from django.utils import timezone
from djblets.features.testing import override_feature_checks
from djblets.webapi.errors import INVALID_ATTRIBUTE, INVALID_FORM_DATA
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.diffviewer.commit_utils import serialize_validation_info
from reviewboard.diffviewer.models import DiffCommit
from reviewboard.reviews.models import ReviewRequestDraft
from reviewboard.webapi.errors import DIFF_EMPTY, DIFF_TOO_BIG
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_extra_data import ExtraDataItemMixin
from reviewboard.webapi.tests.mimetypes import (
    draft_diffcommit_list_mimetype,
    draft_diffcommit_item_mimetype)
from reviewboard.webapi.tests.test_diffcommit import compare_diffcommit
from reviewboard.webapi.tests.urls import (get_draft_diffcommit_item_url,
                                           get_draft_diffcommit_list_url)


class ResourceListTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Tests for DraftDiffCommitResource list resource."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-request/<id>/draft/commits/'
    resource = resources.draft_diffcommit

    compare_item = compare_diffcommit

    _DEFAULT_DIFF_CONTENTS = (
        b'diff --git a/readme b/readme\n'
        b'index d6613f5..5b50866 100644\n'
        b'--- a/readme\n'
        b'+++ b/readme\n'
        b'@@ -1 +1,3 @@\n'
        b' Hello there\n'
        b'+\n'
        b'+Oh hi!\n'
    )

    _DEFAULT_POST_DATA = {
        'author_name': 'Author',
        'author_date': timezone.now().strftime(DiffCommit.ISO_DATE_FORMAT),
        'author_email': 'author@example.com',
        'committer_name': 'Committer',
        'committer_date': timezone.now().strftime(DiffCommit.ISO_DATE_FORMAT),
        'committer_email': 'committer@example.com',
        'commit_message': 'Commit message',
        'commit_id': 'r1',
        'parent_id': 'r0',
    }

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            public=True)
        diffset = self.create_diffset(review_request=review_request)
        return get_draft_diffcommit_list_url(review_request, diffset.revision)

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            public=True)
        diffset = self.create_diffset(review_request=review_request)
        return get_draft_diffcommit_list_url(review_request, diffset.revision)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            publish=True,
            with_local_site=with_local_site,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)
        items = []

        if populate_items:
            items.append(self.create_diffcommit(diffset=diffset,
                                                repository=repository))

        return (
            get_draft_diffcommit_list_url(
                review_request,
                diffset.revision,
                local_site_name=local_site_name),
            draft_diffcommit_list_mimetype,
            items,
        )

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site=False,
                              local_site_name=None, post_valid_data=False):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            with_local_site=with_local_site,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)

        if post_valid_data:
            diff = SimpleUploadedFile('diff', self._DEFAULT_DIFF_CONTENTS)
            post_data = dict(self._DEFAULT_POST_DATA,
                             **{'diff': diff})
        else:
            post_data = {}

        return (
            get_draft_diffcommit_list_url(review_request,
                                          diffset.revision,
                                          local_site_name=local_site_name),
            draft_diffcommit_item_mimetype,
            post_data,
            [],
        )

    def check_post_result(self, user, rsp):
        self.assertIn('draft_commit', rsp)
        item = rsp['draft_commit']
        diffcommit = DiffCommit.objects.get(pk=item['id'])

        self.compare_item(item, diffcommit)

    @webapi_test_template
    def test_post_empty(self):
        """Testing the POST <URL> API with an empty diff"""
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'diff': SimpleUploadedFile('diff', b'     ',
                                               content_type='text/x-patch'),
                }),
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_EMPTY.code)

    @webapi_test_template
    def test_post_too_large(self):
        """Testing the POST <URL> API with a diff that is too large"""
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)

        diff = SimpleUploadedFile('diff',
                                  self._DEFAULT_DIFF_CONTENTS,
                                  content_type='text/x-patch')

        with self.siteconfig_settings({'diffviewer_max_diff_size': 1},
                                      reload_settings=False):
            with override_feature_checks(self.override_features):
                rsp = self.api_post(
                    get_draft_diffcommit_list_url(review_request,
                                                  diffset.revision),
                    dict(self._DEFAULT_POST_DATA, **{
                        'diff': diff,
                    }),
                    expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_TOO_BIG.code)
        self.assertEqual(rsp['max_size'], 1)

    @webapi_test_template
    def test_post_no_history_allowed(self):
        """Testing the POST <URL> API for a review request created without
        history support
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=False)
        ReviewRequestDraft.create(review_request)
        diffset = self.create_diffset(review_request, draft=True)

        diff = SimpleUploadedFile('diff',
                                  self._DEFAULT_DIFF_CONTENTS,
                                  content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'diff': diff,
                }),
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_ATTRIBUTE.code)
        self.assertEqual(
            rsp['reason'],
            'This review request was not created with support for multiple '
            'commits.\n\n'
            'Use the draft_diff resource to upload diffs instead. See the '
            'draft_diff link on the parent resource for the URL.')

    @webapi_test_template
    def test_post_parent_diff(self):
        """Testing the POST <URL> API with a parent diff"""
        parent_diff_contents = (
            b'diff --git a/foo b/foo\n'
            b'new file mode 100644\n'
            b'index 0000000..e69de29\n'
        )

        diff_contents = (
            b'diff --git a/foo b/foo\n'
            b'index e69ded29..03b37a0 100644\n'
            b'--- a/foo\n'
            b'+++ b/foo\n'
            b'@@ -0,0 +1 @@'
            b'+foo bar baz qux\n'
        )
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)

        diff = SimpleUploadedFile('diff', diff_contents,
                                  content_type='text/x-patch')
        parent_diff = SimpleUploadedFile('parent_diff',
                                         parent_diff_contents,
                                         content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'diff': diff,
                    'parent_diff': parent_diff,
                }),
                expected_mimetype=draft_diffcommit_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('draft_commit', rsp)

        item_rsp = rsp['draft_commit']
        self.compare_item(item_rsp, DiffCommit.objects.get(pk=item_rsp['id']))

        commit = DiffCommit.objects.get(pk=item_rsp['id'])
        self.compare_item(item_rsp, commit)

        files = list(commit.files.all())
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertIsNotNone(f.parent_diff)

    @webapi_test_template
    def test_post_parent_diff_subsequent(self):
        """Testing the POST <URL> API with a parent diff on a subsequent commit
        """
        parent_diff_contents = (
            b'diff --git a/foo b/foo\n'
            b'new file mode 100644\n'
            b'index 0000000..e69de29\n'
        )

        diff_contents = (
            b'diff --git a/foo b/foo\n'
            b'index e69ded29..03b37a0 100644\n'
            b'--- a/foo\n'
            b'+++ b/foo\n'
            b'@@ -0,0 +1 @@'
            b'+foo bar baz qux\n'
        )

        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=True)

        diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(repository, diffset)

        diff = SimpleUploadedFile('diff', diff_contents,
                                  content_type='text/x-patch')
        parent_diff = SimpleUploadedFile('parent_diff', parent_diff_contents,
                                         content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'commit_id': 'r0',
                    'parent_id': 'r1',
                    'diff': diff,
                    'parent_diff': parent_diff,
                }),
                expected_mimetype=draft_diffcommit_item_mimetype,
                expected_status=201)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('draft_commit', rsp)

        item_rsp = rsp['draft_commit']
        commit = DiffCommit.objects.get(pk=item_rsp['id'])
        self.compare_item(item_rsp, commit)

        files = list(commit.files.all())
        self.assertEqual(len(files), 1)

        f = files[0]
        self.assertIsNotNone(f.parent_diff)

    @webapi_test_template
    def test_post_invalid_date_format(self):
        """Testing the POST <URL> API with an invalid date format"""
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)

        diff = SimpleUploadedFile('diff',
                                  self._DEFAULT_DIFF_CONTENTS,
                                  content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'commit_id': 'r0',
                    'parent_id': 'r1',
                    'diff': diff,
                    'committer_date': 'Jun 1 1990',
                    'author_date': 'Jun 1 1990',
                }),
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)

        err_fields = rsp['fields']
        self.assertIn('author_date', err_fields)
        self.assertIn('committer_date', err_fields)

        self.assertEqual(err_fields['author_date'],
                         ['This date must be in ISO 8601 format.'])
        self.assertEqual(err_fields['committer_date'],
                         ['This date must be in ISO 8601 format.'])

    @webapi_test_template
    def test_post_subsequent(self):
        """Testing the POST <URL> API with a subsequent commit"""
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user,
            create_with_history=True)
        diffset = self.create_diffset(review_request, draft=True)

        commit = self.create_diffcommit(
            repository,
            diffset,
            diff_contents=self._DEFAULT_DIFF_CONTENTS)

        validation_info = serialize_validation_info({
            commit.commit_id: {
                'parent_id': commit.parent_id,
                'tree': {
                    'added': [],
                    'modified': [{
                        'filename': 'readme',
                        'revision': '5b50866',
                    }],
                    'removed': [],
                },
            },
        })

        diff = SimpleUploadedFile(
            'diff',
            (b'diff --git a/readme b/readme\n'
             b'index 5b50866..f00f00f 100644\n'
             b'--- a/readme\n'
             b'+++ a/readme\n'
             b'@@ -1 +1,4 @@\n'
             b' Hello there\n'
             b' \n'
             b' Oh hi!\n'
             b'+Goodbye!\n'),
            content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'commit_id': 'r2',
                    'parent_id': 'r1',
                    'diff': diff,
                    'validation_info': validation_info,
                }),
                expected_mimetype=draft_diffcommit_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('draft_commit', rsp)

        item_rsp = rsp['draft_commit']
        self.compare_item(item_rsp, DiffCommit.objects.get(pk=item_rsp['id']))

    @webapi_test_template
    def test_post_finalized(self):
        """Testing the POST <URL> API after the parent DiffSet has been
        finalized
        """
        with override_feature_checks(self.override_features):
            review_request = self.create_review_request(
                create_repository=True,
                submitter=self.user,
                create_with_history=True)

            diffset = self.create_diffset(review_request, draft=True)
            diffset.finalize_commit_series(
                cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                validation_info=None,
                validate=False,
                save=True)

            diff = SimpleUploadedFile(
                'diff',
                self._DEFAULT_DIFF_CONTENTS,
                content_type='text/x-patch')

            rsp = self.api_post(
                get_draft_diffcommit_list_url(review_request,
                                              diffset.revision),
                dict(self._DEFAULT_POST_DATA, **{
                    'validation_info': serialize_validation_info({}),
                    'diff': diff,
                }),
                expected_status=400)

            self.assertEqual(rsp, {
                'stat': 'fail',
                'err': {
                    'code': INVALID_ATTRIBUTE.code,
                    'msg': INVALID_ATTRIBUTE.msg,
                    'type': INVALID_ATTRIBUTE.error_type,
                },
                'reason': 'The diff has already been finalized.',
            })


class ResourceItemTests(ExtraDataItemMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Tests for DraftDiffCommitResource item resource."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-request/<id>/draft/commits/<commit-id>/'
    resource = resources.draft_diffcommit

    compare_item = compare_diffcommit

    def setup_http_not_allowed_item_test(self, user):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    submitter=user)
        diffset = self.create_diffset(review_request, draft=True)
        commit = self.create_diffcommit(repository=repository,
                                        diffset=diffset)

        return get_draft_diffcommit_item_url(review_request,
                                             diffset.revision,
                                             commit.commit_id)

    def setup_http_not_allowed_list_test(self, user):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    submitter=user)
        diffset = self.create_diffset(review_request, draft=True)
        commit = self.create_diffcommit(repository=repository,
                                        diffset=diffset)

        return get_draft_diffcommit_item_url(review_request,
                                             diffset.revision,
                                             commit.commit_id)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            with_local_site=with_local_site)
        diffset = self.create_diffset(review_request, draft=True)
        commit = self.create_diffcommit(repository=repository,
                                        diffset=diffset)

        return (
            get_draft_diffcommit_item_url(review_request,
                                          diffset.revision,
                                          commit.commit_id,
                                          local_site_name=local_site_name),
            draft_diffcommit_item_mimetype,
            commit)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            with_local_site=with_local_site)
        diffset = self.create_diffset(review_request, draft=True)
        commit = self.create_diffcommit(repository=repository,
                                        diffset=diffset)

        if put_valid_data:
            request_data = {}
        else:
            request_data = {}

        return (
            get_draft_diffcommit_item_url(review_request,
                                          diffset.revision,
                                          commit.commit_id,
                                          local_site_name=local_site_name),
            draft_diffcommit_item_mimetype,
            request_data,
            commit,
            [])

    def check_put_result(self, user, item_rsp, commit):
        commit = DiffCommit.objects.get(pk=commit.pk)
        self.compare_item(item_rsp, commit)



"""Unit tests for reviewboard.diffviewer.parser."""

from djblets.testing.decorators import add_fixtures

from reviewboard.diffviewer.testing.mixins import DiffParserTestingMixin
from reviewboard.diffviewer.parser import (BaseDiffParser,
                                           DiffParser,
                                           ParsedDiff,
                                           ParsedDiffChange,
                                           ParsedDiffFile)
from reviewboard.testing import TestCase


class ParsedDiffTests(TestCase):
    """Unit tests for reviewboard.diffviewer.parser.ParsedDiff."""

    def test_init(self):
        """Testing ParsedDiff.__init__"""
        parser = BaseDiffParser(data=b'')
        parsed_diff = ParsedDiff(parser=parser)

        self.assertIs(parsed_diff.parser, parser)


class ParsedDiffChangeTests(TestCase):
    """Unit tests for reviewboard.diffviewer.parser.ParsedDiffChange."""

    def test_init(self):
        """Testing ParsedDiffChange.__init__"""
        parser = BaseDiffParser(data=b'')
        parsed_diff = ParsedDiff(parser=parser)
        parsed_diff_change = ParsedDiffChange(parsed_diff=parsed_diff)

        self.assertEqual(parsed_diff.changes, [parsed_diff_change])
        self.assertIs(parsed_diff_change.parent_parsed_diff, parsed_diff)


class ParsedDiffFileTests(TestCase):
    """Unit tests for reviewboard.diffviewer.parser.ParsedDiffFile."""

    def test_init_with_parsed_diff_change(self):
        """Testing ParsedDiffFile.__init__ with parsed_diff_change="""
        parser = BaseDiffParser(data=b'')
        parsed_diff = ParsedDiff(parser=parser)
        parsed_diff_change = ParsedDiffChange(parsed_diff=parsed_diff)
        parsed_diff_file = ParsedDiffFile(
            parsed_diff_change=parsed_diff_change)

        self.assertIs(parsed_diff_file.parser, parser)
        self.assertEqual(parsed_diff_change.files, [parsed_diff_file])
        self.assertEqual(parsed_diff_file.parent_parsed_diff_change,
                         parsed_diff_change)


class DiffParserTest(DiffParserTestingMixin, TestCase):
    """Unit tests for reviewboard.diffviewer.parser.DiffParser."""

    def test_form_feed(self):
        """Testing DiffParser with a form feed in the file"""
        diff = (
            b'--- README  123\n'
            b'+++ README  (new)\n'
            b'@@ -1,4 +1,6 @@\n'
            b' Line 1\n'
            b' Line 2\n'
            b'+\x0c\n'
            b'+Inserted line\n'
            b' Line 3\n'
            b' Line 4\n'
        )

        parsed_files = DiffParser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'README',
            orig_file_details=b'123',
            modified_filename=b'README',
            modified_file_details=b'(new)',
            insert_count=2,
            data=diff)

    def test_line_counts(self):
        """Testing DiffParser with insert/delete line counts"""
        diff = (
            b'+ This is some line before the change\n'
            b'- And another line\n'
            b'Index: foo\n'
            b'- One last.\n'
            b'--- README  123\n'
            b'+++ README  (new)\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'-blah\n'
            b'+blah!\n'
            b'-blah...\n'
            b'+blah?\n'
            b'-blah!\n'
            b'+blah?!\n'
        )

        parsed_files = DiffParser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'README',
            orig_file_details=b'123',
            modified_filename=b'README',
            modified_file_details=b'(new)',
            insert_count=3,
            delete_count=4,
            data=diff)

    @add_fixtures(['test_scmtools'])
    def test_raw_diff_with_diffset(self):
        """Testing DiffParser.raw_diff with DiffSet"""
        repository = self.create_repository(tool_name='Test')
        diffset = self.create_diffset(repository=repository)

        self.create_diffcommit(
            diffset=diffset,
            commit_id='r1',
            parent_id='r0',
            diff_contents=(
                b'diff --git a/ABC b/ABC\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- ABC\n'
                b'+++ ABC\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-line!\n'
                b'+line..\n'
            ))
        self.create_diffcommit(
            diffset=diffset,
            commit_id='r2',
            parent_id='r1',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+Hi, world!\n'
            ))
        self.create_diffcommit(
            diffset=diffset,
            commit_id='r4',
            parent_id='r3',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 197009f..87abad9 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hi, world!\n'
                b'+Yo, world.\n'
            ))

        cumulative_diff = (
            b'diff --git a/ABC b/ABC\n'
            b'index 94bdd3e..197009f 100644\n'
            b'--- ABC\n'
            b'+++ ABC\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-line!\n'
            b'+line..\n'
            b'diff --git a/README b/README\n'
            b'index 94bdd3e..87abad9 100644\n'
            b'--- README\n'
            b'+++ README\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-Hello, world!\n'
            b'+Yo, world.\n'
        )

        diffset.finalize_commit_series(
            cumulative_diff=cumulative_diff,
            validation_info=None,
            validate=False,
            save=True)

        parser = DiffParser(b'')
        self.assertEqual(parser.raw_diff(diffset), cumulative_diff)

    @add_fixtures(['test_scmtools'])
    def test_raw_diff_with_diffcommit(self):
        """Testing DiffParser.raw_diff with DiffCommit"""
        repository = self.create_repository(tool_name='Test')
        diffset = self.create_diffset(repository=repository)

        commit1_diff = (
            b'diff --git a/ABC b/ABC\n'
            b'index 94bdd3e..197009f 100644\n'
            b'--- ABC\n'
            b'+++ ABC\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-line!\n'
            b'+line..\n'
            b'diff --git a/FOO b/FOO\n'
            b'index 84bda3e..b975034 100644\n'
            b'--- FOO\n'
            b'+++ FOO\n'
            b'@@ -1,1 +0,0 @@\n'
            b'-Some line\n'
        )

        commit1 = self.create_diffcommit(
            diffset=diffset,
            commit_id='r1',
            parent_id='r0',
            diff_contents=commit1_diff)
        self.create_diffcommit(
            diffset=diffset,
            commit_id='r2',
            parent_id='r1',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+Hi, world!\n'
            ))
        self.create_diffcommit(
            diffset=diffset,
            commit_id='r4',
            parent_id='r3',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 197009f..87abad9 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hi, world!\n'
                b'+Yo, world.\n'
            ))

        diffset.finalize_commit_series(
            cumulative_diff=(
                b'diff --git a/ABC b/ABC\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- ABC\n'
                b'+++ ABC\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-line!\n'
                b'+line..\n'
                b'diff --git a/FOO b/FOO\n'
                b'index 84bda3e..b975034 100644\n'
                b'--- FOO\n'
                b'+++ FOO\n'
                b'@@ -1,1 +0,0 @@\n'
                b'-Some line\n'
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..87abad9 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+Yo, world.\n'
            ),
            validation_info=None,
            validate=False,
            save=True)

        parser = DiffParser(b'')
        self.assertEqual(parser.raw_diff(commit1), commit1_diff)

    def test_parsed_diff_extra_data(self):
        """Testing custom DiffParser populating a ParsedDiff's extra_data"""
        class CustomParser(DiffParser):
            def parse(self):
                result = super(CustomParser, self).parse()

                self.parsed_diff.extra_data = {
                    'foo': True,
                }

                return result

        diff = self.DEFAULT_FILEDIFF_DATA_DIFF
        parsed_diff_file = CustomParser(diff).parse_diff()

        self.assertEqual(parsed_diff_file.extra_data, {'foo': True})

    def test_parsed_diff_change_extra_data(self):
        """Testing custom DiffParser populating a ParsedDiffChange's
        extra_data
        """
        class CustomParser(DiffParser):
            def parse(self):
                result = super(CustomParser, self).parse()

                self.parsed_diff_change.extra_data = {
                    'foo': True,
                }

                return result

        diff = self.DEFAULT_FILEDIFF_DATA_DIFF
        parsed_diff_file = CustomParser(diff).parse_diff()

        changes = parsed_diff_file.changes
        self.assertEqual(len(changes), 1)
        self.assertEqual(changes[0].extra_data, {'foo': True})

    def test_parsed_diff_file_extra_data(self):
        """Testing custom DiffParser populating a ParsedDiffFile's extra_data
        """
        class CustomParser(DiffParser):
            def parse_diff_header(self, linenum, parsed_file):
                parsed_file.extra_data = {'foo': True}

                return super(CustomParser, self).parse_diff_header(
                    linenum, parsed_file)

        diff = self.DEFAULT_FILEDIFF_DATA_DIFF
        parsed_diff_file = CustomParser(diff).parse_diff()

        changes = parsed_diff_file.changes
        self.assertEqual(len(changes), 1)

        files = changes[0].files
        self.assertEqual(len(files), 1)
        self.assertEqual(files[0].extra_data, {'foo': True})



"""Unit tests for reviewboard.extensions.hooks.AdminWidgetHook."""

from reviewboard.admin.widgets import (BaseAdminWidget,
                                       admin_widgets_registry)
from reviewboard.extensions.hooks import AdminWidgetHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase


class MyAdminWidget(BaseAdminWidget):
    widget_id = 'test-widget'
    name = 'Testing Widget'


class AdminWidgetHookTests(BaseExtensionHookTestCase):
    """Testing AdminWidgetHook."""

    def test_initialize(self):
        """Testing AdminWidgetHook.initialize"""
        AdminWidgetHook(self.extension, MyAdminWidget)

        self.assertIn(MyAdminWidget, admin_widgets_registry)

    def test_shutdown(self):
        """Testing AdminWidgetHook.shutdown"""
        hook = AdminWidgetHook(self.extension, MyAdminWidget)
        hook.disable_hook()

        self.assertNotIn(MyAdminWidget, admin_widgets_registry)



"""Unit tests for reviewboard.attachments.mimetypes.MimetypeHandler.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

import kgb
import mimeparse

from django.contrib.auth.models import User

from reviewboard.attachments.mimetypes import (MimetypeHandler,
                                               register_mimetype_handler,
                                               score_match,
                                               unregister_mimetype_handler)
from reviewboard.attachments.tests.base import BaseFileAttachmentTestCase
from reviewboard.testing import TestCase


class MimetypeTest(MimetypeHandler):
    """Handler for all test mimetypes."""

    supported_mimetypes = ['test/*']


class TestAbcMimetype(MimetypeHandler):
    """Handler for the test/abc mimetype."""

    supported_mimetypes = ['test/abc']


class TestXmlMimetype(MimetypeHandler):
    """Handler for the test/xml mimetype."""

    supported_mimetypes = ['test/xml']


class Test2AbcXmlMimetype(MimetypeHandler):
    """Handler for the test/abc+xml mimetype."""

    supported_mimetypes = ['test2/abc+xml']


class StarDefMimetype(MimetypeHandler):
    """Handler for all /def mimetypes."""

    supported_mimetypes = ['*/def']


class StarAbcDefMimetype(MimetypeHandler):
    """Handler for all /abc+def mimetypes."""

    supported_mimetypes = ['*/abc+def']


class Test3XmlMimetype(MimetypeHandler):
    """Handler for the test3/xml mimetype."""

    supported_mimetypes = ['test3/xml']


class Test3AbcXmlMimetype(MimetypeHandler):
    """Handler for the test3/abc+xml mimetype."""

    supported_mimetypes = ['test3/abc+xml']


class Test3StarMimetype(MimetypeHandler):
    """Handler for all test3 mimetypes."""

    supported_mimetypes = ['test3/*']


class SandboxMimetypeHandler(MimetypeHandler):
    """Handler for image/png mimetypes, used for testing sandboxing."""

    supported_mimetypes = ['image/png']

    def get_icon_url(self):
        """Raise an exception to test sandboxing."""
        raise Exception

    def get_thumbnail(self):
        """Raise an exception to test sandboxing."""
        raise Exception

    def set_thumbnail(self, data):
        """Raise an exception to test sandboxing."""
        raise Exception


class MimetypeHandlerTests(TestCase):
    """Unit tests for reviewboard.attachments.mimetypes.MimetypeHandler.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    def setUp(self):
        """Set up this test case."""
        super(MimetypeHandlerTests, self).setUp()

        # Register test cases in same order as they are defined
        # in this test
        register_mimetype_handler(MimetypeTest)
        register_mimetype_handler(TestAbcMimetype)
        register_mimetype_handler(TestXmlMimetype)
        register_mimetype_handler(Test2AbcXmlMimetype)
        register_mimetype_handler(StarDefMimetype)
        register_mimetype_handler(StarAbcDefMimetype)
        register_mimetype_handler(Test3XmlMimetype)
        register_mimetype_handler(Test3AbcXmlMimetype)
        register_mimetype_handler(Test3StarMimetype)

    def tearDown(self):
        """Tear down this test case."""
        super(MimetypeHandlerTests, self).tearDown()

        # Unregister test cases in same order as they are defined
        # in this test
        unregister_mimetype_handler(MimetypeTest)
        unregister_mimetype_handler(TestAbcMimetype)
        unregister_mimetype_handler(TestXmlMimetype)
        unregister_mimetype_handler(Test2AbcXmlMimetype)
        unregister_mimetype_handler(StarDefMimetype)
        unregister_mimetype_handler(StarAbcDefMimetype)
        unregister_mimetype_handler(Test3XmlMimetype)
        unregister_mimetype_handler(Test3AbcXmlMimetype)
        unregister_mimetype_handler(Test3StarMimetype)

    def _handler_for(self, mimetype):
        mt = mimeparse.parse_mime_type(mimetype)
        score, handler = MimetypeHandler.get_best_handler(mt)
        return handler

    def test_handler_factory(self):
        """Testing matching of factory method for mimetype handlers"""
        # Exact Match
        self.assertEqual(self._handler_for("test/abc"), TestAbcMimetype)
        self.assertEqual(self._handler_for("test2/abc+xml"),
                         Test2AbcXmlMimetype)
        # Handle vendor-specific match
        self.assertEqual(self._handler_for("test/abc+xml"), TestXmlMimetype)
        self.assertEqual(self._handler_for("test2/xml"), Test2AbcXmlMimetype)

    def test_handler_factory_precedence(self):
        """Testing precedence of factory method for mimetype handlers"""
        self.assertEqual(self._handler_for("test2/def"), StarDefMimetype)
        self.assertEqual(self._handler_for("test3/abc+xml"),
                         Test3AbcXmlMimetype)
        self.assertEqual(self._handler_for("test3/xml"), Test3XmlMimetype)
        self.assertEqual(self._handler_for("foo/abc+def"), StarAbcDefMimetype)
        self.assertEqual(self._handler_for("foo/def"), StarDefMimetype)
        # Left match and Wildcard should trump Left Wildcard and match
        self.assertEqual(self._handler_for("test/def"), MimetypeTest)

    def test_mimetype_match_scoring(self):
        """Testing score_match for different mimetype patterns"""
        def assert_score(pattern, test, score):
            self.assertAlmostEqual(
                score_match(mimeparse.parse_mime_type(pattern),
                            mimeparse.parse_mime_type(test)),
                score)

        assert_score('application/reviewboard+x-pdf',
                     'application/reviewboard+x-pdf', 2.0)
        assert_score('application/x-pdf', 'application/x-pdf', 1.9)
        assert_score('text/*', 'text/plain', 1.8)
        assert_score('*/reviewboard+plain', 'text/reviewboard+plain', 1.7)
        assert_score('*/plain', 'text/plain', 1.6)
        assert_score('application/x-javascript', 'application/x-pdf', 0)


class SandboxTests(kgb.SpyAgency, BaseFileAttachmentTestCase):
    """Unit tests for MimetypeHandler sandboxing.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    def setUp(self):
        """Set up this test case."""
        super(SandboxTests, self).setUp()

        register_mimetype_handler(SandboxMimetypeHandler)

        user = User.objects.create_user(username='reviewboard',
                                        password='password',
                                        email='reviewboard@example.com')

        review_request = self.create_review_request(submitter=user)
        self.file_attachment = self.create_file_attachment(
            review_request=review_request)

    def tearDown(self):
        """Tear down this test case."""
        super(SandboxTests, self).tearDown()

        unregister_mimetype_handler(SandboxMimetypeHandler)

    def test_get_thumbnail(self):
        """Testing FileAttachment sandboxes MimetypeHandler.get_thumbnail"""
        self.spy_on(SandboxMimetypeHandler.get_thumbnail,
                    owner=SandboxMimetypeHandler)

        self.file_attachment.thumbnail
        self.assertTrue(SandboxMimetypeHandler.get_thumbnail.called)

    def test_set_thumbnail(self):
        """Testing FileAttachment sandboxes MimetypeHandler.set_thumbnail"""
        self.spy_on(SandboxMimetypeHandler.set_thumbnail,
                    owner=SandboxMimetypeHandler)

        self.file_attachment.thumbnail = None
        self.assertTrue(SandboxMimetypeHandler.set_thumbnail.called)

    def test_get_icon_url(self):
        """Testing FileAttachment sandboxes MimetypeHandler.get_icon_url"""
        self.spy_on(SandboxMimetypeHandler.get_icon_url,
                    owner=SandboxMimetypeHandler)

        self.file_attachment.icon_url
        self.assertTrue(SandboxMimetypeHandler.get_icon_url.called)



"""Registry for Review UIs.

Version Added:
    7.0
"""

from __future__ import annotations

from importlib import import_module
from typing import Iterator, Type

from django.utils.translation import gettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         NOT_REGISTERED)

from reviewboard.registries.registry import Registry
from reviewboard.reviews.ui.base import ReviewUI


class ReviewUIRegistry(Registry[Type[ReviewUI]]):
    """A registry for managing ReviewUIs.

    Version Added:
        7.0
    """

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered Review UI.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s" is not a registered Review UI.'
        ),
    }

    def get_defaults(self) -> Iterator[type[ReviewUI]]:
        """Yield the built-in Review UIs.

        Yields:
            class:
            The built-in Review UI classes.
        """
        builtin_review_uis = (
            ('image', 'ImageReviewUI'),
            ('markdownui', 'MarkdownReviewUI'),
            ('text', 'TextBasedReviewUI'),
        )

        for _module, _cls_name in builtin_review_uis:
            mod = import_module(f'reviewboard.reviews.ui.{_module}')
            yield getattr(mod, _cls_name)



# The oauth2_provider module expects that the app specified in
# settings.OAUTH2_PROVIDER['APPLICATION_MODULE'] has already been added in the
# database. We need to specify this dependency to ensure this app is processed
# first, to fulfill that requirement.
BEFORE_EVOLUTIONS = [
    'oauth2_provider',
]


# We also need to ensure that we've installed reviewboard.site first, since
# otherwise it would apply after oauth2_provider.
AFTER_EVOLUTIONS = [
    'site',
]


SEQUENCE = [
    'disabled_for_security',
    'upgrade_to_1_6_3',
]



"""Views for rendering diff fragments."""

from __future__ import annotations

import io
import logging
import os
import struct
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Tuple

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.core.files.base import ContentFile
from django.db.models import Q
from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.utils.cache import patch_cache_control
from django.utils.safestring import SafeString, mark_safe
from django.views.generic.base import ContextMixin, View
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.http import encode_etag
from djblets.views.generic.etag import ETagViewMixin
from housekeeping import deprecate_non_keyword_only_args
from typing_extensions import TypedDict

from reviewboard.attachments.mimetypes import guess_mimetype
from reviewboard.attachments.models import FileAttachment
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.diffviewer.diffutils import (
    get_file_chunks_in_range,
    get_is_new_file,
    get_last_header_before_line,
    get_last_line_number_in_diff,
    get_sha256,
)
from reviewboard.diffviewer.models import FileDiff
from reviewboard.diffviewer.settings import DiffSettings
from reviewboard.diffviewer.views import (DiffFragmentView,
                                          exception_traceback_string)
from reviewboard.reviews.models import Comment, FileAttachmentComment
from reviewboard.reviews.ui.base import ReviewUI
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin
from reviewboard.scmtools.core import FileLookupContext
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from collections.abc import Mapping

    from reviewboard.diffviewer.diffutils import SerializedDiffFile
    from reviewboard.diffviewer.models import DiffCommit
    from reviewboard.diffviewer.renderers import DiffRenderer


logger = logging.getLogger(__name__)


class CommentFragment(TypedDict):
    """A comment fragment.

    Version Added:
        6.0
    """

    #: The diff comment.
    #:
    #: Type:
    #:     reviewboard.reviews.models.diff_comment.Comment
    comment: Comment

    #: The rendered diff fragment HTML.
    #:
    #: Type:
    #:     django.utils.safestring.SafeString
    html: SafeString

    #: The diff chunks included in the fragment.
    #:
    #: This is the information returned by
    #: :py:func:`~reviewboard.diffviewer.diffutils.get_file_chunks_in_range`.
    #:
    #: Type:
    #:     list
    chunks: List[Dict]


def build_diff_comment_fragments(
    *,
    comments: List[Comment],
    context: Dict[str, Any],
    comment_template_name: str = 'reviews/diff_comment_fragment.html',
    error_template_name: str = 'diffviewer/diff_fragment_error.html',
    lines_of_context: Optional[List[int]] = None,
    show_controls: bool = False,
    request: Optional[HttpRequest] = None,
) -> Tuple[bool, List[CommentFragment]]:
    """Construct and return the comment fragment data.

    Args:
        comments (list of reviewboard.reviews.models.diff_comment.Comment):
            The comments to return diff fragments for.

        context (dict):
            The rendering context.

        comment_template_name (str, optional):
            The template to use for rendering the comment fragment.

        error_template_name (str, optional):
            The template to use for rendering errors.

        lines_of_context (list of int, optional):
            A 2-element list containing the number of additional lines of
            context to render above and below each fragment.

        show_controls (bool, optional):
            Whether to show expand controls in the rendered output.

        request (django.http.HttpRequest, optional):
            The HTTP request from the client, if available.

    Returns:
        tuple:
        A 2-tuple containing:

        Tuple:
            0 (bool):
                Whether any errors occurred.

            1 (list of CommentFragment):
                The rendered comment fragments.
    """
    comment_entries = []
    had_error = False
    siteconfig = SiteConfiguration.objects.get_current()
    diff_settings = DiffSettings.create(request=request)

    if lines_of_context is None:
        lines_of_context = [0, 0]

    for comment in comments:
        try:
            base_commit: Optional[DiffCommit] = None
            tip_commit: Optional[DiffCommit] = None

            base_filediff = comment.base_filediff

            if base_filediff:
                base_commit = base_filediff.commit

            if comment.filediff.commit_id:
                tip_commit = comment.filediff.commit

            max_line = get_last_line_number_in_diff(
                context=context,
                filediff=comment.filediff,
                interfilediff=comment.interfilediff,
                diff_settings=diff_settings,
                base_filediff=base_filediff,
                base_commit=base_commit,
                tip_commit=tip_commit)

            first_line = max(1, comment.first_line - lines_of_context[0])
            last_line = min(comment.last_line + lines_of_context[1], max_line)
            num_lines = last_line - first_line + 1

            chunks = list(get_file_chunks_in_range(
                context=context,
                filediff=comment.filediff,
                interfilediff=comment.interfilediff,
                first_line=first_line,
                num_lines=num_lines,
                diff_settings=diff_settings,
                base_filediff=base_filediff,
                base_commit=base_commit,
                tip_commit=tip_commit
            ))

            comment_context = {
                'comment': comment,
                'header': get_last_header_before_line(
                    context=context,
                    filediff=comment.filediff,
                    interfilediff=comment.interfilediff,
                    target_line=first_line,
                    diff_settings=diff_settings,
                    base_filediff=base_filediff,
                    base_commit=base_commit,
                    tip_commit=tip_commit),
                'chunks': chunks,
                'domain': Site.objects.get_current().domain,
                'domain_method': siteconfig.get('site_domain_method'),
                'lines_of_context': lines_of_context,
                'expandable_above': show_controls and first_line != 1,
                'expandable_below': show_controls and last_line != max_line,
                'is_new_file': get_is_new_file(
                    filediff=comment.filediff,
                    interfilediff=comment.interfilediff,
                    base_filediff=base_filediff),
                'collapsible': lines_of_context != [0, 0],
                'lines_above': first_line - 1,
                'lines_below': max_line - last_line,
                'first_line': first_line,
            }
            comment_context.update(context)
            content = render_to_string(template_name=comment_template_name,
                                       context=comment_context,
                                       request=request)
        except Exception as e:
            content = exception_traceback_string(
                None, e, error_template_name, {
                    'comment': comment,
                    'file': {
                        'orig_filename': comment.filediff.source_file,
                        'index': None,
                        'filediff': comment.filediff,
                    },
                    'domain': Site.objects.get_current().domain,
                    'domain_method': siteconfig.get("site_domain_method"),
                })

            # It's bad that we failed, and we'll return a 500, but we'll
            # still return content for anything we have. This will prevent any
            # caching.
            had_error = True
            chunks = []

        comment_entries.append(CommentFragment(
            comment=comment,
            html=content,
            chunks=chunks))

    return had_error, comment_entries


class CommentDiffFragmentsView(ReviewRequestViewMixin, ETagViewMixin,
                               ContextMixin, View):
    """View for rendering a section of a diff that a comment pertains to.

    This takes in one or more
    :py:class:`~reviewboard.reviews.models.diff_comment.Comment` IDs
    (comma-separated) as part of the URL and returns a payload containing
    data and HTML for each comment's diff fragment, which the client can
    parse in order to dynamically load the fragments into the page.

    The resulting format is a custom, condensed format containing the comment
    ID and HTML for each diff fragment. It's designed to be quick to parse and
    reduces the amount of data to send across the wire (unlike a format like
    JSON, which would add overhead to the serialization/deserialization time
    and data size when storing HTML, or JavaScript, which releases prior to
    3.0 used to handle injecting fragments into the DOM).

    Each entry in the payload is in the following format, with all entries
    joined together:

        <comment ID>\\n
        <html length>\\n
        <html content>

    The format is subject to change without notice, and should not be relied
    upon by third parties.

    The following URL query options are supported:

    ``allow_expansion``:
        Whether expansion controls should be shown to the user. To enable
        this, the caller must pass a value of ``1``. This is disabled by
        default.

    ``lines_of_context``:
        The number of lines of context before and after the commented region
        of the diff. This is in the form of ``pre,post``, where both are the
        numbers of lines. This defaults to ``0,0``.
    """

    comment_template_name = 'reviews/diff_comment_fragment.html'
    error_template_name = 'diffviewer/diff_fragment_error.html'

    content_type = 'application/javascript'

    EXPIRATION_SECONDS = 60 * 60 * 24 * 365  # 1 year

    def get_etag_data(
        self,
        request: HttpRequest,
        comment_ids: str,
        *args,
        **kwargs,
    ) -> str:
        """Return an ETag for the view.

        This will look up state needed for the request and generate a
        suitable ETag. Some of the information will be stored for later
        computation of the template context.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            comment_ids (str):
                A list of comment IDs to render.

            *args (tuple, unused):
                Positional arguments passed to the handler.

            **kwargs (dict, unused):
                Keyword arguments passed to the handler.

        Returns:
            str:
            The ETag for the page.

        Raises:
            django.http.Http404:
                The given parameters were not valid.
        """
        q = (Q(pk__in=comment_ids.split(',')) &
             Q(review__review_request=self.review_request))

        if request.user.is_authenticated:
            q &= Q(review__public=True) | Q(review__user=request.user)
        else:
            q &= Q(review__public=True)

        comments = list(Comment.objects.filter(q).order_by('pk'))

        if not comments:
            raise Http404()

        timestamps = ':'.join(
            comment.timestamp.isoformat()
            for comment in comments
        )

        self.comments = comments

        return f'{comment_ids}:{timestamps}:{settings.TEMPLATE_SERIAL}'

    def get(
        self,
        request: HttpRequest,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        This will generate a payload for the diff comments being loaded and
        pass them in a format that can be parsed by the client.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            **kwargs (dict):
                Keyword arguments passed to the view.

        Returns:
            django.http.HttpResponse:
            The HTTP response containing the fragments payload.
        """
        lines_of_context = request.GET.get('lines_of_context', '0,0')
        allow_expansion = (request.GET.get('allow_expansion') == '1')

        try:
            lines_of_context = [int(i) for i in lines_of_context.split(',')]

            # Ensure that we have 2 values for lines_of_context. If only one is
            # given, assume it is both the before and after context. If more
            # than two are given, only consider the first two. If somehow we
            # get no lines of context value, we will default to [0, 0].

            if len(lines_of_context) == 1:
                lines_of_context.append(lines_of_context[0])
            elif len(lines_of_context) > 2:
                lines_of_context = lines_of_context[0:2]
            elif len(lines_of_context) == 0:
                raise ValueError
        except ValueError:
            lines_of_context = [0, 0]

        context = super().get_context_data(**kwargs)
        context.update({
            'request': request,
            'user': request.user,
        })

        payload = io.BytesIO()
        had_error, comment_entries = build_diff_comment_fragments(
            comments=self.comments,
            context=context,
            comment_template_name=self.comment_template_name,
            error_template_name=self.error_template_name,
            lines_of_context=lines_of_context,
            show_controls=allow_expansion)

        for entry in comment_entries:
            html = entry['html'].strip().encode('utf-8')

            payload.write(struct.pack(b'<LL', entry['comment'].pk, len(html)))
            payload.write(html)

        result = payload.getvalue()
        payload.close()

        response = HttpResponse(result,
                                content_type='text/plain; charset=utf-8')

        if had_error:
            patch_cache_control(response,
                                no_cache=True,
                                no_store=True,
                                max_age=0,
                                must_revalidate=True)

        return response


class ReviewsDiffFragmentView(ReviewRequestViewMixin, DiffFragmentView):
    """Renders a fragment from a file in the diff viewer.

    Displays just a fragment of a diff or interdiff owned by the given
    review request. The fragment is identified by the chunk index in the
    diff.

    ``review_request_id``:
        The ID of the ReviewRequest containing the diff to render.

    ``revision``:
        The DiffSet revision to render.

    ``filediff_id``:
        The ID of the FileDiff within the DiffSet.

    The following may also be provided:

    ``interdiff_revision``:
        The second DiffSet revision in an interdiff revision range.

    ``chunk_index``:
        The index (0-based) of the chunk to render. If left out, the
        entire file will be rendered.

    ``local_site``:
        The LocalSite the ReviewRequest must be on, if any.

    See :py:class:`~reviewboard.diffviewer.views.DiffFragmentView` for the
    accepted query parameters.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the view.

        Args:
            *args (tuple):
                Positional arguments to pass through to the parent class.

            **kwargs (dict):
                Keyword arguments to pass through to the parent class.
        """
        super().__init__(*args, **kwargs)

        self._cached_diffset_info: (Mapping[str, Any] | None) = None

    def process_diffset_info(
        self,
        revision: int,
        interdiff_revision: (int | None) = None,
        **kwargs,
    ) -> Mapping[str, Any]:
        """Process and return information on the desired diff.

        The diff IDs and other data passed to the view can be processed and
        converted into DiffSets. A dictionary with the DiffSet and FileDiff
        information will be returned.

        Args:
            revision (int):
                The revision of the diff to view.

            interdiff_revision (int, optional):
                The second diff revision if viewing an interdiff.

            **kwargs (dict):
                Keyword arguments passed to the view.

        Returns:
            dict:
            Information on the diff for use in the template and in queries.
        """
        if self._cached_diffset_info is not None:
            return self._cached_diffset_info

        draft = self.review_request.get_draft(user=self.request.user)

        if interdiff_revision is not None:
            interdiffset = self.get_diff(interdiff_revision, draft)
        else:
            interdiffset = None

        diffset = self.get_diff(revision, draft)

        info = super().process_diffset_info(
            diffset_or_id=diffset,
            interdiffset_or_id=interdiffset,
            **kwargs)

        self._cached_diffset_info = info

        return info

    def make_etag(
        self,
        *,
        request: (HttpRequest | None) = None,
        **kwargs,
    ) -> str:
        """Return an ETag identifying this render.

        Version Added:
            8.0

        Args:
            request (django.http.HttpRequest):
                The request from the client.

                Version Added:
                    8.0

            **kwargs (dict):
                Additional keyword arguments passed to the function.

        Returns:
            str:
            The encoded ETag identifying this render.

        Raises:
            django.http.Http404:
                The diff for the given parameters was not found.
        """
        etag = super().make_etag(request=request, **kwargs)

        filediff_id = kwargs.get('filediff_id')
        filediff = get_object_or_404(FileDiff, pk=filediff_id)

        if filediff.binary:
            # For binary files, serialized comments get included with the
            # Review UI's rendered HTML. We therefore need to include the
            # comment timestamps in the ETag or reloading the page can
            # show out of date comments.
            #
            # This isn't an issue for non-binary files because all the comments
            # are included along with the main diff viewer page, rather than
            # with individual fragments.
            assert request is not None

            diff_info = self.process_diffset_info(
                base_filediff_id=request.GET.get('base-filediff-id'),
                **kwargs)
            diff_file = diff_info['diff_file']

            orig_attachment, modified_attachment = \
                self._get_attachment_objects_for_binary(
                    request=request,
                    filediff=diff_file['filediff'],
                    interfilediff=diff_file['interfilediff'],
                    base_filediff=diff_file['base_filediff'],
                    force_interdiff=diff_file['force_interdiff'],
                    is_new_file=diff_file['is_new_file'],
                )

            if orig_attachment or modified_attachment:
                comment_timestamps = (
                    FileAttachmentComment.objects.for_file_attachment(
                        attachment=modified_attachment,
                        diff_against_file_attachment=orig_attachment,
                        user=request.user,
                    )
                    .order_by('pk')
                    .values_list('timestamp', flat=True)
                )

                if comment_timestamps:
                    timestamps = ':'.join(
                        timestamp.isoformat()
                        for timestamp in comment_timestamps
                    )
                    etag = encode_etag(f'{etag}:{timestamps}')

        return etag

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def create_renderer(
        self,
        *,
        context: Mapping[str, Any],
        renderer_settings: Mapping[str, Any],
        diff_file: SerializedDiffFile,
        request: HttpRequest,
        **kwargs,
    ) -> DiffRenderer:
        """Create the DiffRenderer for this fragment.

        This will augment the renderer for binary files by looking up
        file attachments, if review UIs are involved, disabling caching.

        Version Changed:
            8.0:
            * Deprecated non-keyword arguments.
            * Added the ``request`` parameter.

        Args:
            context (dict):
                The current render context.

            renderer_settings (dict):
                The diff renderer settings.

            diff_file (reviewboard.diffviewer.diffutils.SerializedDiffFile):
                The information on the diff file to render.

            request (django.http.HttpRequest):
                The request from the client.

            **kwargs (dict):
                Additional keyword arguments from the parent class.

        Returns:
            reviewboard.diffviewer.renderers.DiffRenderer:
            The resulting diff renderer.
        """
        renderer = super().create_renderer(
            request=request,
            context=context,
            renderer_settings=renderer_settings,
            diff_file=diff_file,
            **kwargs)

        if diff_file['binary']:
            # Determine the file attachments to display in the diff viewer,
            # if any.
            orig_attachment, modified_attachment = \
                self._get_attachment_objects_for_binary(
                    request=request,
                    filediff=diff_file['filediff'],
                    interfilediff=diff_file['interfilediff'],
                    base_filediff=diff_file['base_filediff'],
                    force_interdiff=diff_file['force_interdiff'],
                    is_new_file=diff_file['is_new_file'],
                )

            diff_review_ui_html: (str | None) = None
            orig_review_ui_class: (type[ReviewUI[Any, Any, Any]] | None) = None
            orig_review_ui_html: (str | None) = None
            modified_review_ui_class: (
                type[ReviewUI[Any, Any, Any]] |
                None
            ) = None
            modified_review_ui_html: (str | None) = None
            review_request = context['review_request']

            if orig_attachment:
                orig_review_ui_class = ReviewUI.for_object(orig_attachment)

            if modified_attachment:
                modified_review_ui_class = ReviewUI.for_object(
                    modified_attachment)

            if (orig_review_ui_class is not None and
                orig_review_ui_class is modified_review_ui_class and
                orig_review_ui_class.supports_diffing):
                review_ui = orig_review_ui_class(
                    review_request=review_request,
                    obj=modified_attachment)
                review_ui.set_diff_against(orig_attachment)
                diff_review_ui_html = self._render_review_ui(review_ui)
            else:
                if orig_review_ui_class:
                    review_ui = orig_review_ui_class(
                        review_request=review_request,
                        obj=orig_attachment)
                    orig_review_ui_html = self._render_review_ui(review_ui)

                if modified_review_ui_class:
                    review_ui = modified_review_ui_class(
                        review_request=review_request,
                        obj=modified_attachment)
                    modified_review_ui_html = self._render_review_ui(review_ui)

            if (diff_review_ui_html or orig_review_ui_html or
                modified_review_ui_html):
                # Don't cache the view, because the Review UI may care about
                # state that we can't anticipate. At the least, it may have
                # comments or other data that change between renders, and we
                # don't want that to go stale.
                renderer.allow_caching = False

            renderer.extra_context.update({
                'orig_diff_file_attachment': orig_attachment,
                'modified_diff_file_attachment': modified_attachment,
                'orig_attachment_review_ui_html': orig_review_ui_html,
                'modified_attachment_review_ui_html': modified_review_ui_html,
                'diff_attachment_review_ui_html': diff_review_ui_html,
            })

        renderer.extra_context.update(
            self._get_download_links(renderer, diff_file))

        return renderer

    def _get_attachment_objects_for_binary(
        self,
        *,
        request: HttpRequest,
        filediff: FileDiff,
        interfilediff: FileDiff | None,
        base_filediff: FileDiff | None,
        force_interdiff: bool,
        is_new_file: bool,
    ) -> tuple[FileAttachment | None, FileAttachment | None]:
        """Return the file attachments to display in the diff viewer.

        For any binary files which are part of the change, we use file
        attachments for storage and review.

        Version Added:
            8.0

        Args:
            request (django.http.HttpRequest):
                The request from the client.

            filediff (reviewboard.diffviewer.models.FileDiff):
                The filediff for the file.

            interfilediff (reviewboard.diffviewer.models.FileDiff):
                The filediff for the interdiff, if present.

            base_filediff (reviewboard.diffviewer.models.FileDiff):
                The filediff for the base diff, if present.

            force_interdiff (bool):
                Whether to force rendering an interdiff.

                This is used to show correct interdiffs for files that were
                reverted in later versions.

            is_new_file (bool):
                Whether the filediff corresponds to a newly-added file.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (reviewboard.attachments.models.FileAttachment):
                    The file attachment for the original version of the file,
                    if present.

                1 (reviewboard.attachments.models.FileAttachment):
                    The file attachment for the modified version of the file,
                    if present.
        """
        orig_attachment: (FileAttachment | None) = None
        modified_attachment: (FileAttachment | None) = None

        if force_interdiff:
            if interfilediff:
                orig_attachment = self._get_diff_file_attachment(
                    filediff=filediff)
                modified_attachment = self._get_diff_file_attachment(
                    filediff=interfilediff)
            else:
                # We're forcing an interdiff but have no interfilediff, which
                # means this is a reverted file.
                orig_attachment = self._get_diff_file_attachment(
                    filediff=filediff)
                modified_attachment = self._get_diff_file_attachment(
                    filediff=filediff,
                    use_modified=False)
        else:
            modified_attachment = self._get_diff_file_attachment(
                filediff=filediff)

            if base_filediff is not None:
                orig_attachment = self._get_diff_file_attachment(
                    filediff=base_filediff)
            elif not is_new_file:
                orig_attachment = self._get_diff_file_attachment(
                    filediff=filediff, use_modified=False)

                if (orig_attachment is None and
                    modified_attachment is not None):
                    # We only fetch the original version of the file if we
                    # already have an attachment for the modified version. This
                    # way we're not cluttering up the DB and filesystem with
                    # attachments that aren't helpful for the review process.
                    orig_attachment = self._create_attachment_for_orig(
                        request=request, filediff=filediff)

        return orig_attachment, modified_attachment

    def get_context_data(
        self,
        **kwargs,
    ) -> Dict[str, Any]:
        """"Return context for rendering the view.

        Args:
            **kwargs (dict):
                Keyword arguments passed to the view.

        Returns:
            dict:
            Context to use for rendering templates.
        """
        return {
            'review_request': self.review_request,
        }

    def _get_download_links(
        self,
        renderer: DiffRenderer,
        diff_file: Dict[str, Any],
    ) -> Dict[str, Optional[str]]:
        """Return links for downloading the files used for the diff.

        Args:
            renderer (reviewboard.diffviewer.renderers.DiffRenderer):
                The diff renderer.

            diff_file (dict):
                The diff file information.

        Returns:
            dict:
            A dictionary containing ``download_orig_url`` and
            ``download_modified_url`` keys, with the URLs to download the
            original and modified files.
        """
        if diff_file['binary']:
            orig_attachment = \
                renderer.extra_context['orig_diff_file_attachment']
            modified_attachment = \
                renderer.extra_context['modified_diff_file_attachment']

            if orig_attachment:
                download_orig_url = orig_attachment.get_absolute_url()
            else:
                download_orig_url = None

            if modified_attachment:
                download_modified_url = modified_attachment.get_absolute_url()
            else:
                download_modified_url = None
        else:
            filediff = diff_file['filediff']
            interfilediff = diff_file['interfilediff']
            diffset = filediff.diffset

            if interfilediff:
                orig_url_name = 'download-modified-file'
                modified_revision = interfilediff.diffset.revision
                modified_filediff_id = interfilediff.pk
            else:
                orig_url_name = 'download-orig-file'
                modified_revision = diffset.revision
                modified_filediff_id = filediff.pk

            filediff_is_empty = filediff.is_diff_empty

            if interfilediff:
                interfilediff_is_empty = interfilediff.is_diff_empty
            else:
                interfilediff_is_empty = True

            if (filediff_is_empty and
                interfilediff_is_empty and
                (diff_file['newfile'] or diff_file['deleted'])):
                # If an empty file was added or removed, we don't want
                # to show download links.
                download_orig_url = None
                download_modified_url = None
            else:
                download_orig_url = local_site_reverse(
                    orig_url_name,
                    request=self.request,
                    kwargs={
                        'review_request_id': self.review_request.display_id,
                        'revision': diffset.revision,
                        'filediff_id': filediff.pk,
                    })

                download_modified_url = local_site_reverse(
                    'download-modified-file',
                    request=self.request,
                    kwargs={
                        'review_request_id': self.review_request.display_id,
                        'revision': modified_revision,
                        'filediff_id': modified_filediff_id,
                    })

        return {
            'download_orig_url': download_orig_url,
            'download_modified_url': download_modified_url,
        }

    def _render_review_ui(
        self,
        review_ui: Optional[ReviewUI],
        inline_only: bool = True,
    ) -> Optional[SafeString]:
        """Render the review UI for a file attachment.

        Args:
            review_ui (reviewboard.reviews.ui.base.ReviewUI):
                The review UI to render.

            inline_only (bool):
                Whether to limit rendering to review UIs that support inline
                mode.

        Returns:
            django.utils.safestring.SafeString:
            The rendered review UI HTML, if available. ``None`` if not.
        """
        if review_ui and (not inline_only or review_ui.allow_inline):
            return mark_safe(review_ui.render_to_string(self.request,
                                                        inline=True))

        return None

    def _get_diff_file_attachment(
        self,
        *,
        filediff: FileDiff,
        use_modified: bool = True,
    ) -> Optional[FileAttachment]:
        """Fetch the FileAttachment associated with a FileDiff.

        Args:
            filediff (reviewboard.diffviewer.models.filediff.FileDiff):
                The FileDiff to find a linked attachment.

            use_modified (bool):
                Whether to return the attachment for the modified version (new
                file). If ``False``, will return the FileAttachment for the
                original file.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The matching file attachment, if available. If no matching
            attachment is found, or if more than one is found associated with
            the FileDiff, ``None`` is returned.
        """
        if not filediff:
            return None

        try:
            return FileAttachment.objects.get_for_filediff(filediff,
                                                           use_modified)
        except ObjectDoesNotExist:
            return None
        except MultipleObjectsReturned:
            # Only one FileAttachment should be associated with a FileDiff
            logger.exception('More than one FileAttachments associated with '
                             'FileDiff %s',
                             filediff.pk,
                             extra={'request': self.request})

            return None

    def _create_attachment_for_orig(
        self,
        *,
        request: HttpRequest,
        filediff: FileDiff,
    ) -> FileAttachment | None:
        """Create an attachment for the original version of a binary file.

        New versions of binary files in diffs are expected to be uploaded by
        RBTools, but the original version may need to be fetched from the
        repository.

        Version Added:
            7.0

        Args:
            request (django.http.HttpRequest):
                The request from the server.

            filediff (reviewboard.diffviewer.models.filediff.Filediff):
                The FileDiff for the attachment.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The newly-created file attachment, or ``None`` if the creation
            failed.
        """
        extra_data = filediff.extra_data or {}

        if filediff.commit is not None:
            commit_extra_data = filediff.commit.extra_data
        else:
            commit_extra_data = {}

        diffset = filediff.diffset

        context = FileLookupContext(
            request=request,
            base_commit_id=diffset.base_commit_id,
            diff_extra_data=diffset.extra_data,
            commit_extra_data=commit_extra_data,
            file_extra_data=extra_data)

        try:
            repository = filediff.get_repository()
            file_contents = repository.get_file(
                path=filediff.source_file,
                revision=filediff.source_revision,
                context=context)
            attachment_extra_data = {
                'sha256_checksum': get_sha256(file_contents),
            }

            with ContentFile(file_contents) as file_obj:
                attachment = FileAttachment.objects.create_from_filediff(
                    extra_data=attachment_extra_data,
                    filediff=filediff,
                    from_modified=False,
                    mimetype=guess_mimetype(file_obj))

                attachment.file.save(
                    os.path.basename(filediff.source_file),
                    file_obj)

            logger.debug('Creating new file attachment for binary file %s '
                         '(%s) in repository %s',
                         filediff.source_file,
                         filediff.source_revision,
                         repository)

            return attachment
        except Exception as e:
            logger.exception(
                'Unable to fetch original version of binary file %s (%s): %s',
                filediff.source_file,
                filediff.source_revision,
                e)

            return None



"""Unit tests for reviewboard.extensions.hooks.HostingServiceHook."""

from reviewboard.extensions.hooks import HostingServiceHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase
from reviewboard.hostingsvcs.base import (BaseHostingService,
                                          hosting_service_registry)
from reviewboard.scmtools.errors import FileNotFoundError


class TestService(BaseHostingService):
    hosting_service_id = 'test-service'
    name = 'Test Service'

    def get_file(self, repository, path, revision, *args, **kwargs):
        """Return the specified file from the repository.

        If the given file path is ``/invalid-path``, the file will be assumed
        to not exist and
        :py:exc:`reviewboard.scmtools.errors.FileNotFoundError` will be raised.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the file belongs to.

            path (unicode):
                The file path.

            revision (unicode):
                The file revision.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            unicode: The file data.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                Raised if the file does not exist.
        """
        if path == '/invalid-path':
            raise FileNotFoundError(path, revision)

        return super(TestService, self).get_file(repository, path, revision,
                                                 *args, **kwargs)

    def get_file_exists(self, repository, path, revision, *args, **kwargs):
        """Return the specified file from the repository.

        If the given file path is ``/invalid-path``, the file will
        be assumed to not exist.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the file belongs to.

            path (unicode):
                The file path.

            revision (unicode):
                The file revision.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool: Whether or not the file exists.
        """
        if path == '/invalid-path':
            return False

        return super(TestService, self).get_file_exists(
            repository, path, revision, *args, **kwargs)


class HostingServiceHookTests(BaseExtensionHookTestCase):
    """Testing HostingServiceHook."""

    def test_register(self):
        """Testing HostingServiceHook initializing"""
        HostingServiceHook(self.extension, TestService)

        self.assertEqual(
            hosting_service_registry.get_hosting_service('test-service'),
            TestService)

    def test_register_without_hosting_service_id(self):
        """Testing HostingServiceHook initializing without hosting_service_id
        """
        class TestServiceWithoutID(TestService):
            hosting_service_id = None

        message = 'TestServiceWithoutID.hosting_service_id must be set.'

        with self.assertRaisesMessage(ValueError, message):
            HostingServiceHook(self.extension, TestServiceWithoutID)

    def test_unregister(self):
        """Testing HostingServiceHook uninitializing"""
        hook = HostingServiceHook(self.extension, TestService)
        hook.disable_hook()

        self.assertIsNone(
            hosting_service_registry.get_hosting_service('test-service'))



"""Unit tests for reviewboard.themes.ui.base.BaseUITheme.

Version Added:
    7.0
"""

from __future__ import annotations

from reviewboard.testing import TestCase
from reviewboard.themes.ui.base import BaseUITheme, InkColorScheme


class BaseUIThemeTests(TestCase):
    """Unit tests for BaseUITheme.

    Version Added:
        7.0
    """

    def test_get_html_attrs(self) -> None:
        """Testing BaseUITheme.get_html_attrs"""
        class MyUITheme(BaseUITheme):
            theme_id = 'test'
            name = 'Test'
            ink_color_scheme = InkColorScheme.SYSTEM

        self.assertEqual(
            MyUITheme().get_html_attrs(request=self.create_http_request()),
            {
                'data-ink-color-scheme': 'system',
                'data-theme': 'test',
            })



"""Unit tests for BaseHostingServiceAuthForm."""

from __future__ import annotations

from typing import TYPE_CHECKING

import kgb

from reviewboard.certs.cert import Certificate, CertificateFingerprints
from reviewboard.certs.errors import (CertificateVerificationError,
                                      CertificateVerificationFailureCode)
from reviewboard.certs.manager import cert_manager
from reviewboard.certs.tests.testcases import (CaptureSSLMixin,
                                               TEST_SHA256,
                                               TEST_TRUST_CERT_PEM)
from reviewboard.deprecation import (RemovedInReviewBoard10_0Warning,
                                     RemovedInReviewBoard90Warning)
from reviewboard.hostingsvcs.base import hosting_service_registry
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            TwoFactorAuthCodeRequiredError)
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceAuthForm
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.scmtools.certs import Certificate as LegacyCertificate
from reviewboard.scmtools.errors import UnverifiedCertificateError
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase
from reviewboard.testing.hosting_services import (SelfHostedTestService,
                                                  TestService)

if TYPE_CHECKING:
    from djblets.testing.testcases import ExpectedWarning


class HostingServiceAuthFormTests(kgb.SpyAgency, CaptureSSLMixin, TestCase):
    """Unit tests for BaseHostingServiceAuthForm."""

    fixtures = ['test_scmtools']

    def setUp(self) -> None:
        """Set up state for the test.

        This will clear out the certs directory before running a test.
        """
        super().setUp()

        hosting_service_registry.register(TestService)
        hosting_service_registry.register(SelfHostedTestService)

    def tearDown(self) -> None:
        """Tear down state for the test.

        This will clear out the certs directory after running a test.
        """
        hosting_service_registry.unregister(SelfHostedTestService)
        hosting_service_registry.unregister(TestService)

        super().tearDown()

    def test_override_help_texts(self):
        """Testing BaseHostingServiceAuthForm subclasses overriding help texts
        """
        class MyAuthForm(BaseHostingServiceAuthForm):
            class Meta:
                help_texts = {
                    'hosting_account_username': 'My help text.',
                }

        form = MyAuthForm(hosting_service_cls=TestService)

        self.assertEqual(form.fields['hosting_account_username'].help_text,
                         'My help text.')

    def test_override_labels(self):
        """Testing BaseHostingServiceAuthForm subclasses overriding labels"""
        class MyAuthForm(BaseHostingServiceAuthForm):
            class Meta:
                labels = {
                    'hosting_account_username': 'My label.',
                }

        form = MyAuthForm(hosting_service_cls=TestService)

        self.assertEqual(form.fields['hosting_account_username'].label,
                         'My label.')

    def test_get_credentials_default(self):
        """Testing BaseHostingServiceAuthForm.get_credentials default behavior
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.get_credentials(),
            {
                'username': 'myuser',
                'password': 'mypass',
            })

    def test_get_credentials_default_with_2fa_code(self):
        """Testing BaseHostingServiceAuthForm.get_credentials default behavior
        with two-factor auth code
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
                'hosting_account_two_factor_auth_code': '123456',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.get_credentials(),
            {
                'username': 'myuser',
                'password': 'mypass',
                'two_factor_auth_code': '123456',
            })

    def test_get_credentials_with_form_prefix(self):
        """Testing BaseHostingServiceAuthForm.get_credentials default behavior
        with form prefix
        """
        form = BaseHostingServiceAuthForm(
            {
                'myservice-hosting_account_username': 'myuser',
                'myservice-hosting_account_password': 'mypass',
                'myservice-hosting_account_two_factor_auth_code': '123456',
            },
            hosting_service_cls=TestService,
            prefix='myservice')

        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.get_credentials(),
            {
                'username': 'myuser',
                'password': 'mypass',
                'two_factor_auth_code': '123456',
            })

    def test_save_new_account(self):
        """Testing BaseHostingServiceAuthForm.save with new account"""
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()

        self.assertIsNotNone(hosting_account.pk)
        self.assertEqual(hosting_account.service_name, 'test')
        self.assertEqual(hosting_account.username, 'myuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertIsNone(hosting_account.hosting_url)
        self.assertIsNone(hosting_account.local_site)

    def test_save_new_account_with_existing_stored(self):
        """Testing BaseHostingServiceAuthForm.save with new account matching
        existing stored account information
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())

        orig_account = HostingServiceAccount.objects.create(
            service_name='test',
            username='myuser')

        hosting_account = form.save()

        self.assertIsNotNone(hosting_account.pk)
        self.assertEqual(hosting_account.pk, orig_account.pk)
        self.assertEqual(hosting_account.service_name, 'test')
        self.assertEqual(hosting_account.username, 'myuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertIsNone(hosting_account.hosting_url)
        self.assertIsNone(hosting_account.local_site)

    def test_save_new_account_with_hosting_url(self):
        """Testing BaseHostingServiceAuthForm.save with new account and
        hosting URL
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
                'hosting_url': 'example.com',
            },
            hosting_service_cls=SelfHostedTestService)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()

        self.assertIsNotNone(hosting_account.pk)
        self.assertEqual(hosting_account.service_name, 'self_hosted_test')
        self.assertEqual(hosting_account.username, 'myuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertEqual(hosting_account.hosting_url, 'example.com')
        self.assertIsNone(hosting_account.local_site)

    def test_save_new_account_with_hosting_url_not_self_hosted(self):
        """Testing BaseHostingServiceAuthForm.save with new account and
        hosting URL with non-self-hosted service
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
                'hosting_url': 'example.com',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())
        self.assertNotIn('hosting_url', form.cleaned_data)

        hosting_account = form.save()
        self.assertIsNone(hosting_account.hosting_url)

    def test_save_new_account_without_hosting_url_self_hosted(self):
        """Testing BaseHostingServiceAuthForm.save with new account and no
        hosting URL with a self-hosted service
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=SelfHostedTestService)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'hosting_url': ['This field is required.'],
            })

    def test_save_new_account_with_local_site(self):
        """Testing BaseHostingServiceAuthForm.save with new account and Local
        Site
        """
        local_site = LocalSite.objects.create(name='test-site')
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
            local_site=local_site)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()

        self.assertIsNotNone(hosting_account.pk)
        self.assertEqual(hosting_account.service_name, 'test')
        self.assertEqual(hosting_account.username, 'myuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertEqual(hosting_account.local_site, local_site)
        self.assertIsNone(hosting_account.hosting_url)

    def test_save_new_account_without_username(self):
        """Testing BaseHostingServiceAuthForm.save with new account and no
        username in credentials
        """
        class MyAuthForm(BaseHostingServiceAuthForm):
            def get_credentials(self):
                return {}

        form = MyAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())

        expected_message = (
            'Hosting service implementation error: '
            'MyAuthForm.get_credentials() must return a "username" key.'
        )

        with self.assertRaisesMessage(AuthorizationError, expected_message):
            form.save()

    def test_save_existing_account(self):
        """Testing BaseHostingServiceAuthForm.save with updating existing
        account
        """
        orig_account = HostingServiceAccount.objects.create(
            service_name='test',
            username='myuser')

        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
            hosting_account=orig_account)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()

        self.assertIs(hosting_account, orig_account)
        self.assertEqual(hosting_account.pk, orig_account.pk)
        self.assertEqual(hosting_account.service_name, 'test')
        self.assertEqual(hosting_account.username, 'myuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertIsNone(hosting_account.hosting_url)
        self.assertIsNone(hosting_account.local_site)

    def test_save_existing_account_new_username(self):
        """Testing BaseHostingServiceAuthForm.save with updating existing
        account with new username
        """
        orig_account = HostingServiceAccount.objects.create(
            service_name='test',
            username='myuser')

        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'mynewuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
            hosting_account=orig_account)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()

        self.assertIs(hosting_account, orig_account)
        self.assertEqual(hosting_account.pk, orig_account.pk)
        self.assertEqual(hosting_account.service_name, 'test')
        self.assertEqual(hosting_account.username, 'mynewuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertIsNone(hosting_account.hosting_url)
        self.assertIsNone(hosting_account.local_site)

    def test_save_existing_account_new_hosting_url(self):
        """Testing BaseHostingServiceAuthForm.save with updating existing
        account with new hosting URL
        """
        orig_account = HostingServiceAccount.objects.create(
            service_name='self_hosted_test',
            username='myuser',
            hosting_url='example1.com')

        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
                'hosting_url': 'example2.com',
            },
            hosting_service_cls=SelfHostedTestService,
            hosting_account=orig_account)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()

        self.assertIs(hosting_account, orig_account)
        self.assertEqual(hosting_account.pk, orig_account.pk)
        self.assertEqual(hosting_account.service_name, 'self_hosted_test')
        self.assertEqual(hosting_account.username, 'myuser')
        self.assertEqual(hosting_account.data['password'], 'mypass')
        self.assertEqual(hosting_account.hosting_url, 'example2.com')
        self.assertIsNone(hosting_account.local_site)

    def test_save_existing_account_new_service_fails(self):
        """Testing BaseHostingServiceAuthForm.save with updating existing
        account with new hosting service fails
        """
        orig_account = HostingServiceAccount.objects.create(
            service_name='self_hosted_test',
            username='myuser',
            hosting_url='example1.com')

        expected_message = (
            'This account is not compatible with this hosting service '
            'configuration.'
        )

        with self.assertRaisesMessage(ValueError, expected_message):
            BaseHostingServiceAuthForm(hosting_service_cls=TestService,
                                   hosting_account=orig_account)

    def test_save_existing_account_new_local_site_fails(self):
        """Testing BaseHostingServiceAuthForm.save with updating existing
        account with new Local Site fails
        """
        orig_account = HostingServiceAccount.objects.create(
            service_name='text',
            username='myuser')

        expected_message = (
            'This account is not compatible with this hosting service '
            'configuration.'
        )

        with self.assertRaisesMessage(ValueError, expected_message):
            BaseHostingServiceAuthForm(
                hosting_service_cls=TestService,
                hosting_account=orig_account,
                local_site=LocalSite.objects.create(name='test-site'))

    def test_save_with_2fa_code_required(self):
        """Testing BaseHostingServiceAuthForm.save with two-factor auth code
        required
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': '2fa-user',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())

        self.assertFalse(
            form.fields['hosting_account_two_factor_auth_code'].required)

        with self.assertRaises(TwoFactorAuthCodeRequiredError):
            form.save()

        self.assertTrue(
            form.fields['hosting_account_two_factor_auth_code'].required)

    def test_save_with_2fa_code_provided(self):
        """Testing BaseHostingServiceAuthForm.save with two-factor auth code
        provided
        """
        form = BaseHostingServiceAuthForm(
            {
                'hosting_account_username': '2fa-user',
                'hosting_account_password': 'mypass',
                'hosting_account_two_factor_auth_code': '123456',
            },
            hosting_service_cls=TestService)

        self.assertTrue(form.is_valid())

        hosting_account = form.save()
        self.assertEqual(hosting_account.service_name, 'test')
        self.assertEqual(hosting_account.username, '2fa-user')

    def test_save_with_cert_error(self) -> None:
        """Testing BaseHostingServiceAuthForm.save with
        UnverifiedCertificateError
        """
        fingerprints = CertificateFingerprints(sha256=TEST_SHA256)

        class _MyAuthForm(BaseHostingServiceAuthForm):
            def authorize(self, *args, **kwargs) -> None:
                verified = cert_manager.is_certificate_verified(
                    hostname='example.com',
                    port=443,
                    latest_fingerprints=fingerprints,
                )

                if not verified:
                    raise CertificateVerificationError(
                        code=CertificateVerificationFailureCode.NOT_TRUSTED,
                        certificate=Certificate(
                            hostname='example.com',
                            port=443,
                            fingerprints=fingerprints,
                            cert_data=TEST_TRUST_CERT_PEM,
                        ),
                    )

        form = _MyAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
        )
        self.spy_on(form.authorize)

        self.assertTrue(form.is_valid())

        message = (
            'The SSL certificate provided by example.com has not been '
            'signed by a trusted Certificate Authority and may not be safe. '
            'The certificate needs to be verified in Review Board before '
            'the server can be accessed. Certificate details: '
            'hostname="example.com", port=443, issuer="example.com", '
            'fingerprint=SHA256=79:19:70:AE:A6:1B:EB:BC:35:7C:B8:54:B1:6A:'
            'AD:79:FF:F7:28:69:02:5E:C3:6F:B3:C2:B4:FD:84:66:DF:8F'
        )

        with self.assertRaisesMessage(CertificateVerificationError, message):
            form.save(allow_authorize=True)

        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=fingerprints,
        ))
        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
        ))
        self.assertSpyCallCount(form.authorize, 1)

    def test_save_with_cert_error_and_trust_host(self) -> None:
        """Testing BaseHostingServiceAuthForm.save with
        CertificateVerificationError and trust_host=True
        """
        fingerprints = CertificateFingerprints(sha256=TEST_SHA256)

        class _MyAuthForm(BaseHostingServiceAuthForm):
            def authorize(self, *args, **kwargs) -> None:
                verified = cert_manager.is_certificate_verified(
                    hostname='example.com',
                    port=443,
                    latest_fingerprints=fingerprints,
                )

                if not verified:
                    raise CertificateVerificationError(
                        code=CertificateVerificationFailureCode.NOT_TRUSTED,
                        certificate=Certificate(
                            hostname='example.com',
                            port=443,
                            fingerprints=fingerprints,
                            cert_data=TEST_TRUST_CERT_PEM,
                        ),
                    )

        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=fingerprints,
        ))

        form = _MyAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
        )
        self.spy_on(form.authorize)

        self.assertTrue(form.is_valid())

        form.save(allow_authorize=True,
                  trust_host=True)

        self.assertTrue(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=fingerprints,
        ))
        self.assertAttrsEqual(
            cert_manager.get_certificate(
                hostname='example.com',
                port=443,
            ),
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'port': 443,
            })
        self.assertSpyCallCount(form.authorize, 2)

    def test_save_with_legacy_cert_error(self) -> None:
        """Testing BaseHostingServiceAuthForm.save with
        legacy UnverifiedCertificateError
        """
        fingerprints = CertificateFingerprints(sha256=TEST_SHA256)

        class _MyAuthForm(BaseHostingServiceAuthForm):
            def authorize(self, *args, **kwargs) -> None:
                verified = cert_manager.is_certificate_verified(
                    hostname='example.com',
                    port=443,
                    latest_fingerprints=fingerprints,
                )

                if not verified:
                    raise UnverifiedCertificateError(
                        certificate=LegacyCertificate(
                            pem_data=TEST_TRUST_CERT_PEM.decode('utf-8'),
                            issuer='issuer',
                            hostname='example.com',
                            fingerprint=TEST_SHA256.replace(':', '').lower(),
                        )
                    )

        form = _MyAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
        )
        self.spy_on(form.authorize)

        self.assertTrue(form.is_valid())

        warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'UnverifiedCertificateError is deprecated in favor of '
                    'reviewboard.certs.errors.CertificateVerificationError, '
                    'and will be removed in Review Board 9.'
                ),
            },
        ]

        message = (
            'The SSL certificate for this repository (hostname '
            '"example.com", fingerprint "791970aea61bebbc357cb854b16aad79f'
            'ff72869025ec36fb3c2b4fd8466df8f") was not verified and might '
            'not be safe. This certificate needs to be verified before the '
            'repository can be accessed.'
        )

        with self.assertWarnings(warnings):
            with self.assertRaisesMessage(UnverifiedCertificateError, message):
                form.save(allow_authorize=True)

        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=fingerprints,
        ))
        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
        ))
        self.assertSpyCallCount(form.authorize, 1)

    def test_save_with_legacy_cert_error_and_trust_host(self) -> None:
        """Testing BaseHostingServiceAuthForm.save with
        legacy UnverifiedCertificateError and trust_host=True
        """
        fingerprints = CertificateFingerprints(sha256=TEST_SHA256)

        class _MyAuthForm(BaseHostingServiceAuthForm):
            def authorize(self, *args, **kwargs) -> None:
                verified = cert_manager.is_certificate_verified(
                    hostname='example.com',
                    port=443,
                    latest_fingerprints=fingerprints,
                )

                if not verified:
                    raise UnverifiedCertificateError(
                        certificate=LegacyCertificate(
                            pem_data=TEST_TRUST_CERT_PEM.decode('utf-8'),
                            issuer='issuer',
                            hostname='example.com',
                            fingerprint=TEST_SHA256.replace(':', '').lower(),
                        )
                    )

        self.assertFalse(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=fingerprints,
        ))

        form = _MyAuthForm(
            {
                'hosting_account_username': 'myuser',
                'hosting_account_password': 'mypass',
            },
            hosting_service_cls=TestService,
        )
        self.spy_on(form.authorize)

        self.assertTrue(form.is_valid())

        warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'UnverifiedCertificateError is deprecated in favor of '
                    'reviewboard.certs.errors.CertificateVerificationError, '
                    'and will be removed in Review Board 9.'
                ),
            },
            {
                'cls': RemovedInReviewBoard10_0Warning,
                'message': (
                    'HostingServiceAccount.accept_certificate() is '
                    'deprecated and will be removed in Review Board 10. '
                    'Use cert_manager.add_certificate() instead.'
                ),
            },
        ]

        with self.assertWarnings(warnings):
            form.save(allow_authorize=True,
                      trust_host=True)

        self.assertTrue(cert_manager.is_certificate_verified(
            hostname='example.com',
            port=443,
            latest_fingerprints=fingerprints,
        ))
        self.assertAttrsEqual(
            cert_manager.get_certificate(
                hostname='example.com',
                port=443,
            ),
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'port': 443,
            })
        self.assertSpyCallCount(form.authorize, 2)



"""Settings for SAML SSO.

Version Added:
    5.0
"""

from django.conf import settings
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from djblets.siteconfig.models import SiteConfiguration
try:
    from onelogin.saml2.constants import OneLogin_Saml2_Constants as constants
except ImportError:
    constants = None

from reviewboard.admin.server import build_server_url


class SAMLSignatureAlgorithm(object):
    """Definitions for the signature algorithm.

    Version Added:
        5.0
    """

    DSA_SHA1 = 'dsa-sha1'
    RSA_SHA1 = 'rsa-sha1'
    RSA_SHA256 = 'rsa-sha256'
    RSA_SHA384 = 'rsa-sha384'
    RSA_SHA512 = 'rsa-sha512'

    CHOICES = (
        (DSA_SHA1, 'DSA-SHA1'),
        (RSA_SHA1, 'RSA-SHA1'),
        (RSA_SHA256, 'RSA-SHA256'),
        (RSA_SHA384, 'RSA-SHA384'),
        (RSA_SHA512, 'RSA-SHA512'),
    )

    if constants:
        TO_SAML2_SETTING_MAP = {
            DSA_SHA1: constants.DSA_SHA1,
            RSA_SHA1: constants.RSA_SHA1,
            RSA_SHA256: constants.RSA_SHA256,
            RSA_SHA384: constants.RSA_SHA384,
            RSA_SHA512: constants.RSA_SHA512,
        }

        FROM_SAML2_SETTING_MAP = {
            constants.DSA_SHA1: DSA_SHA1,
            constants.RSA_SHA1: RSA_SHA1,
            constants.RSA_SHA256: RSA_SHA256,
            constants.RSA_SHA384: RSA_SHA384,
            constants.RSA_SHA512: RSA_SHA512,
        }
    else:
        TO_SAML2_SETTING_MAP = {}
        FROM_SAML2_SETTING_MAP = {}


class SAMLDigestAlgorithm(object):
    """Definitions for the digest algorithm.

    Version Added:
        5.0
    """

    SHA1 = 'sha1'
    SHA256 = 'sha256'
    SHA384 = 'sha384'
    SHA512 = 'sha512'

    CHOICES = (
        (SHA1, 'SHA1'),
        (SHA256, 'SHA256'),
        (SHA384, 'SHA384'),
        (SHA512, 'SHA512'),
    )

    if constants:
        TO_SAML2_SETTING_MAP = {
            SHA1: constants.SHA1,
            SHA256: constants.SHA256,
            SHA384: constants.SHA384,
            SHA512: constants.SHA512,
        }

        FROM_SAML2_SETTING_MAP = {
            constants.SHA1: SHA1,
            constants.SHA256: SHA256,
            constants.SHA384: SHA384,
            constants.SHA512: SHA512,
        }
    else:
        TO_SAML2_SETTING_MAP = {}
        FROM_SAML2_SETTING_MAP = {}


class SAMLBinding(object):
    """Definitions for the binding type.

    Version Added:
        5.0
    """

    HTTP_POST = 'http-post'
    HTTP_REDIRECT = 'http-redirect'

    CHOICES = (
        (HTTP_POST, _('HTTP POST')),
        (HTTP_REDIRECT, _('HTTP Redirect')),
    )

    if constants:
        TO_SAML2_SETTING_MAP = {
            HTTP_POST: constants.BINDING_HTTP_POST,
            HTTP_REDIRECT: constants.BINDING_HTTP_REDIRECT,
        }

        FROM_SAML2_SETTING_MAP = {
            constants.BINDING_HTTP_POST: HTTP_POST,
            constants.BINDING_HTTP_REDIRECT: HTTP_REDIRECT,
        }
    else:
        TO_SAML2_SETTING_MAP = {}
        FROM_SAML2_SETTING_MAP = {}


class SAMLNameIDFormat:
    """Definitions for the NameID format.

    Version Added:
        6.0
    """

    EMAIL = 'email'
    ENTITY = 'entity'
    TRANSIENT = 'transient'
    PERSISTENT = 'persistent'
    UNSPECIFIED = 'unspecified'

    CHOICES = (
        (EMAIL, _('E-mail address')),
        (ENTITY, _('Entity')),
        (TRANSIENT, _('Transient identifier')),
        (PERSISTENT, _('Persistent identifier')),
        (UNSPECIFIED, _('Unspecified')),
    )

    if constants:
        TO_SAML2_SETTING_MAP = {
            EMAIL: constants.NAMEID_EMAIL_ADDRESS,
            ENTITY: constants.NAMEID_ENTITY,
            TRANSIENT: constants.NAMEID_TRANSIENT,
            PERSISTENT: constants.NAMEID_PERSISTENT,
            UNSPECIFIED: constants.NAMEID_UNSPECIFIED,
        }

        FROM_SAML2_SETTING_MAP = {
            constants.NAMEID_EMAIL_ADDRESS: EMAIL,
            constants.NAMEID_ENTITY: ENTITY,
            constants.NAMEID_TRANSIENT: TRANSIENT,
            constants.NAMEID_PERSISTENT: PERSISTENT,
            constants.NAMEID_UNSPECIFIED: UNSPECIFIED,
        }
    else:
        TO_SAML2_SETTING_MAP = {}
        FROM_SAML2_SETTING_MAP = {}


DEFAULT_ATTR_EMAIL = 'User.email'
DEFAULT_ATTR_FIRSTNAME = 'User.FirstName'
DEFAULT_ATTR_FULLNAME = 'User.FullName'
DEFAULT_ATTR_LASTNAME = 'User.LastName'


def get_saml2_settings():
    """Return the SAML2.0 settings.

    Version Added:
        5.0

    Returns:
        dict:
        A dictionary of the settings to use for SAML operations.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    assert constants is not None

    saml_settings = {
        'strict': True,
        'debug': True,
        'idp': {
            'entityId': siteconfig.get('saml_issuer'),
            'singleSignOnService': {
                'url': siteconfig.get('saml_sso_url'),
                'binding': SAMLBinding.TO_SAML2_SETTING_MAP[
                    siteconfig.get('saml_sso_binding_type')],
            },
            'singleLogoutService': {
                'url': siteconfig.get('saml_slo_url'),
                'binding': SAMLBinding.TO_SAML2_SETTING_MAP[
                    siteconfig.get('saml_slo_binding_type')],
            },
            'x509cert': siteconfig.get('saml_verification_cert'),
        },
        'sp': {
            'entityId': build_server_url(
                reverse('sso:saml:metadata', kwargs={'backend_id': 'saml'})),
            'assertionConsumerService': {
                'url': build_server_url(
                    reverse('sso:saml:acs', kwargs={'backend_id': 'saml'})),
                'binding': constants.BINDING_HTTP_POST,
            },
            'singleLogoutService': {
                'url': build_server_url(
                    reverse('sso:saml:sls', kwargs={'backend_id': 'saml'})),
                'binding': constants.BINDING_HTTP_REDIRECT,
            },
            'NameIDFormat': SAMLNameIDFormat.TO_SAML2_SETTING_MAP[
                siteconfig.get('saml_nameid_format')],
            'x509cert': '',
            'privateKey': '',
        },
    }

    if hasattr(settings, 'SAML_CONFIG_ADVANCED'):
        saml_settings.update(settings.SAML_CONFIG_ADVANCED)

    return saml_settings



"""Manager for FileAttachment objects."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Manager, Q

from reviewboard.diffviewer.commit_utils import exclude_ancestor_filediffs
from reviewboard.diffviewer.models import FileDiff

if TYPE_CHECKING:
    from reviewboard.attachments.models import FileAttachment


class FileAttachmentManager(Manager['FileAttachment']):
    """Manages FileAttachment objects.

    Adds utility functions for looking up FileAttachments based on other
    objects.
    """

    def create_from_filediff(
        self,
        filediff: FileDiff,
        from_modified: bool = True,
        save: bool = True,
        **kwargs,
    ) -> FileAttachment:
        """Create a new FileAttachment for a FileDiff.

        FileAttachments created from a FileDiff are used to represent changes
        to binary files which would otherwise not be displayed with the diff.

        An individual FileAttachment can represent either the original or
        modified copy of the file. If 'from_modified' is True, then the
        FileAttachment will be created using the information (filename,
        revision, etc.) for the modified version. If it is False, the
        FileAttachment will be created using the information for the original
        version.

        Args:
            filediff (reviewboard.diffviewer.models.filediff.FileDiff):
                The FileDiff to create the attachment for.

            from_modified (bool, optional):
                Whether to create an attachment for the modified version of the
                file.

            save (bool, optional):
                Whether to save the new object before returning.

            **kwargs (dict):
                Additional keyword arguments to pass through to the
                FileAttachment model.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The newly-created file attachment.
        """
        review_request = filediff.get_review_request()
        assert review_request is not None

        if 'local_site' not in kwargs:
            kwargs['local_site'] = review_request.local_site

        if 'extra_data' not in kwargs:
            extra_data = {}
        else:
            extra_data = kwargs['extra_data'].copy()

        extra_data[self.model.DEFINED_LOCAL_SITE_KEY] = True
        kwargs['extra_data'] = extra_data

        if filediff.is_new or from_modified:
            assert from_modified

            attachment = self.model(
                added_in_filediff=filediff,
                **kwargs)
        else:
            attachment = self.model(
                repo_path=filediff.source_file,
                repo_revision=filediff.source_revision,
                repository=filediff.get_repository(),
                **kwargs)

        attachment._review_request = review_request

        if save:
            attachment.save()

            review_request.file_attachments.add(attachment)

            # If there's a draft, we also need to add it there, otherwise it
            # will disappear as soon as the draft is published.
            draft = review_request.get_draft()

            if draft:
                draft.file_attachments.add(attachment)

        return attachment

    def filter_for_repository(self, repository):
        """Filter results for those on a given repository."""
        return self.filter(
            Q(repository=repository) |
            Q(added_in_filediff__diffset__repository=repository))

    def get_for_filediff(
        self,
        filediff: FileDiff,
        modified: bool = True,
    ) -> Optional[FileAttachment]:
        """Return the FileAttachment matching a FileDiff.

        The FileAttachment associated with the path, revision and repository
        matching the DiffSet will be returned, if it exists.

        It is up to the caller to check for errors.

        Args:
            filediff (reviewboard.diffviewer.models.FileDiff):
                The FileDiff to get the attachment for.

            modified (bool, optional):
                If ``True``, return the FileDiff corresponding to the modified
                revision. If ``False``, return the FileDiff corresponding to
                the original revision.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The attachment for the given FileDiff, if one exists.
        """
        if filediff.is_new and not modified:
            # We never care about the "original" version of a new file.
            return None
        elif filediff.is_new or modified:
            try:
                return self.get(added_in_filediff=filediff)
            except ObjectDoesNotExist:
                pass

            # In the case of review requests created with commit history, we
            # can have multiple FileDiffs for the same file--one as part of the
            # commit, and one as part of the cumulative diff. If we're loading
            # the cumulative diff, we need to look up the attachment
            # corresponding with the latest commit.
            diffset = filediff.diffset

            if not diffset.commit_count:
                return None

            commit_filediffs = list(FileDiff.objects.filter(
                diffset_id=diffset.pk,
                commit__isnull=False,
                dest_file=filediff.dest_file))

            commit_filediffs = exclude_ancestor_filediffs(commit_filediffs)

            if len(commit_filediffs) == 1:
                try:
                    return self.get(added_in_filediff=commit_filediffs[0])
                except ObjectDoesNotExist:
                    pass
        else:
            review_request = filediff.get_review_request()
            assert review_request is not None

            try:
                return self.get(review_request=review_request,
                                repo_path=filediff.source_file,
                                repo_revision=filediff.source_revision,
                                repository=filediff.get_repository())
            except ObjectDoesNotExist:
                pass

        return None



"""Packaging support for Review Board extensions.

Version Changed:
    8.0:
    This has been split into sub-modules: Consumers should update any older
    imports for this module to instead import from
    :py:mod:`reviewboard.extensions.packaging.setuptools_backend` or
    :py:mod:`reviewboard.extensions.packaging.static_media`.
"""

# NOTE: It's important not to import anything directly in this module that
#       depends on a configured Django environment.

from __future__ import annotations

from housekeeping import ClassMovedMixin, func_moved

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
import reviewboard.extensions.packaging.setuptools_backend as \
    setuptools_backend
import reviewboard.extensions.packaging.static_media as static_media


class RBStaticMediaBuildContext(ClassMovedMixin,
                                static_media.RBStaticMediaBuildContext,
                                warning_cls=RemovedInReviewBoard10_0Warning):
    """Context for performing a static media build for an extension.

    This will set up some default NPM workspace paths needed for building
    Review Board extensions.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.extensions.packaging.
        setuptools_backend.RBStaticMediaBuildContext`. The legacy import will
        be removed in Review Board 10.

    Version Added:
        7.0
    """


class RBStaticMediaBuilder(ClassMovedMixin,
                           static_media.RBStaticMediaBuilder,
                           warning_cls=RemovedInReviewBoard10_0Warning):
    """Static media builder for Review Board extensions.

    This will take care of building static media files. As part of this, it
    will set up Babel, Rollup, and TypeScript configuration files needed to
    compile against Review Board JavaScript/TypeScript modules.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.extensions.packaging.
        setuptools_backend.RBStaticMediaBuilder`. The legacy import will be
        removed in Review Board 10.

    Version Added:
        7.0
    """


class BuildStaticFiles(ClassMovedMixin,
                       setuptools_backend.BuildStaticFiles,
                       warning_cls=RemovedInReviewBoard10_0Warning):
    """Builds static files for the extension.

    This will build the static media files used by the extension. JavaScript
    bundles will be minified and versioned. CSS bundles will be processed
    through LessCSS (if using :file:`.less` files), minified and versioned.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.extensions.packaging.
        setuptools_backend.RBStaticMediaBuilder`. The legacy import will be
        removed in Review Board 10.

    Version Added:
        7.0
    """


@func_moved(RemovedInReviewBoard10_0Warning,
            new_func=setuptools_backend.setup)
def setup(**setup_kwargs):
    """Build an extension package.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.extensions.packaging.
        setuptools_backend.setup`. The legacy import will be removed in
        Review Board 10.

    Args:
        **setup_kwargs (dict):
            Keyword arguments to pass to the main setup method.
    """
    return setuptools_backend.setup(**setup_kwargs)



from reviewboard.diffviewer.myersdiff import MyersDiffer
from reviewboard.testing import TestCase


class InterestingLinesTest(TestCase):
    """Unit tests for interesting lines scanner in differ."""

    def test_csharp(self):
        """Testing interesting lines scanner with a C# file"""
        a = ('public class HelloWorld {\n'
             '    public static void Main() {\n'
             '        System.Console.WriteLine("Hello world!");\n'
             '    }\n'
             '}\n')

        b = ('/*\n'
             ' * The Hello World class.\n'
             ' */\n'
             'public class HelloWorld\n'
             '{\n'
             '    /*\n'
             '     * The main function in this class.\n'
             '     */\n'
             '    public static void Main()\n'
             '    {\n'
             '        /*\n'
             '         * Print "Hello world!" to the screen.\n'
             '         */\n'
             '        System.Console.WriteLine("Hello world!");\n'
             '    }\n'
             '}\n')

        lines = self._get_lines(a, b, 'helloworld.cs')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (0, 'public class HelloWorld {\n'),
                (1, '    public static void Main() {\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (3, 'public class HelloWorld\n'),
                (8, '    public static void Main()\n'),
            ])

    def test_java(self):
        """Testing interesting lines scanner with a Java file"""
        a = ('class HelloWorld {\n'
             '    public static void main(String[] args) {\n'
             '        System.out.println("Hello world!");\n'
             '    }\n'
             '}\n')

        b = ('/*\n'
             ' * The Hello World class.\n'
             ' */\n'
             'class HelloWorld\n'
             '{\n'
             '    /*\n'
             '     * The main function in this class.\n'
             '     */\n'
             '    public static void main(String[] args)\n'
             '    {\n'
             '        /*\n'
             '         * Print "Hello world!" to the screen.\n'
             '         */\n'
             '        System.out.println("Hello world!");\n'
             '    }\n'
             '}\n')

        lines = self._get_lines(a, b, 'helloworld.java')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (0, 'class HelloWorld {\n'),
                (1, '    public static void main(String[] args) {\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (3, 'class HelloWorld\n'),
                (8, '    public static void main(String[] args)\n'),
            ])

    def test_javascript(self):
        """Testing interesting lines scanner with a JavaScript file"""
        a = ('function helloWorld() {\n'
             '    alert("Hello world!");\n'
             '}\n'
             '\n'
             'var data = {\n'
             '    helloWorld2: function() {\n'
             '        alert("Hello world!");\n'
             '    }\n'
             '}\n'
             '\n'
             'var helloWorld3 = function() {\n'
             '    alert("Hello world!");\n'
             '}\n')

        b = ('/*\n'
             ' * Prints "Hello world!"\n'
             ' */\n'
             'function helloWorld()\n'
             '{\n'
             '    alert("Hello world!");\n'
             '}\n'
             '\n'
             'var data = {\n'
             '    /*\n'
             '     * Prints "Hello world!"\n'
             '     */\n'
             '    helloWorld2: function()\n'
             '    {\n'
             '        alert("Hello world!");\n'
             '    }\n'
             '}\n'
             '\n'
             'var helloWorld3 = function()\n'
             '{\n'
             '    alert("Hello world!");\n'
             '}\n')

        lines = self._get_lines(a, b, 'helloworld.js')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (0, 'function helloWorld() {\n'),
                (5, '    helloWorld2: function() {\n'),
                (10, 'var helloWorld3 = function() {\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (3, 'function helloWorld()\n'),
                (12, '    helloWorld2: function()\n'),
                (18, 'var helloWorld3 = function()\n'),
            ])

    def test_objective_c(self):
        """Testing interesting lines scanner with an Objective C file"""
        a = ('@interface MyClass : Object\n'
             '- (void) sayHello;\n'
             '@end\n'
             '\n'
             '@implementation MyClass\n'
             '- (void) sayHello {\n'
             '    printf("Hello world!");\n'
             '}\n'
             '@end\n')

        b = ('@interface MyClass : Object\n'
             '- (void) sayHello;\n'
             '@end\n'
             '\n'
             '@implementation MyClass\n'
             '/*\n'
             ' * Prints Hello world!\n'
             ' */\n'
             '- (void) sayHello\n'
             '{\n'
             '    printf("Hello world!");\n'
             '}\n'
             '@end\n')

        lines = self._get_lines(a, b, 'helloworld.m')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (0, '@interface MyClass : Object\n'),
                (4, '@implementation MyClass\n'),
                (5, '- (void) sayHello {\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (0, '@interface MyClass : Object\n'),
                (4, '@implementation MyClass\n'),
                (8, '- (void) sayHello\n'),
            ])

    def test_perl(self):
        """Testing interesting lines scanner with a Perl file"""
        a = ('sub helloWorld {\n'
             '    print "Hello world!"\n'
             '}\n')

        b = ('# Prints Hello World\n'
             'sub helloWorld\n'
             '{\n'
             '    print "Hello world!"\n'
             '}\n')

        lines = self._get_lines(a, b, 'helloworld.pl')

        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], [(0, 'sub helloWorld {\n')])
        self.assertEqual(lines[1], [(1, 'sub helloWorld\n')])

    def test_php(self):
        """Testing interesting lines scanner with a PHP file"""
        a = ('<?php\n'
             'class HelloWorld {\n'
             '    function helloWorld() {\n'
             '        print "Hello world!";\n'
             '    }\n'
             '}\n'
             '?>\n')

        b = ('<?php\n'
             '/*\n'
             ' * Hello World class\n'
             ' */\n'
             'class HelloWorld\n'
             '{\n'
             '    /*\n'
             '     * Prints Hello World\n'
             '     */\n'
             '    function helloWorld()\n'
             '    {\n'
             '        print "Hello world!";\n'
             '    }\n'
             '\n'
             '    public function foo() {\n'
             '        print "Hello world!";\n'
             '    }\n'
             '}\n'
             '?>\n')

        lines = self._get_lines(a, b, 'helloworld.php')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (1, 'class HelloWorld {\n'),
                (2, '    function helloWorld() {\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (4, 'class HelloWorld\n'),
                (9, '    function helloWorld()\n'),
                (14, '    public function foo() {\n'),
            ])

    def test_python(self):
        """Testing interesting lines scanner with a Python file"""
        a = ('class HelloWorld:\n'
             '    def main(self):\n'
             '        print "Hello World"\n')

        b = ('class HelloWorld:\n'
             '    """The Hello World class"""\n'
             '\n'
             '    def main(self):\n'
             '        """The main function in this class."""\n'
             '\n'
             '        # Prints "Hello world!" to the screen.\n'
             '        print "Hello world!"\n')

        lines = self._get_lines(a, b, 'helloworld.py')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (0, 'class HelloWorld:\n'),
                (1, '    def main(self):\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (0, 'class HelloWorld:\n'),
                (3, '    def main(self):\n'),
            ])

    def test_ruby(self):
        """Testing interesting lines scanner with a Ruby file"""
        a = ('class HelloWorld\n'
             '    def helloWorld\n'
             '        puts "Hello world!"\n'
             '    end\n'
             'end\n')

        b = ('# Hello World class\n'
             'class HelloWorld\n'
             '    # Prints Hello World\n'
             '    def helloWorld()\n'
             '        puts "Hello world!"\n'
             '    end\n'
             'end\n')

        lines = self._get_lines(a, b, 'helloworld.rb')

        self.assertEqual(len(lines), 2)
        self.assertEqual(
            lines[0],
            [
                (0, 'class HelloWorld\n'),
                (1, '    def helloWorld\n'),
            ])
        self.assertEqual(
            lines[1],
            [
                (1, 'class HelloWorld\n'),
                (3, '    def helloWorld()\n'),
            ])

    def _get_lines(self, a, b, filename):
        differ = MyersDiffer(a.splitlines(True), b.splitlines(True))
        differ.add_interesting_lines_for_headers(filename)

        # Begin the scan.
        list(differ.get_opcodes())

        return (differ.get_interesting_lines('header', False),
                differ.get_interesting_lines('header', True))



"""Unit tests for reviewboard.accounts.user_details.UserBadge.

Version Added:
    8.0
"""

from __future__ import annotations

from reviewboard.accounts.user_details import UserBadge
from reviewboard.testing import TestCase


class UserBadgeTests(TestCase):
    """Unit tests for UserBadge.

    Version Added:
        8.0
    """

    def test_render_to_string(self) -> None:
        """Testing UserBadge.render_to_string"""
        user = self.create_user()
        badge = UserBadge(user=user,
                          label='My Badge')

        self.assertHTMLEqual(
            badge.render_to_string(),
            '<span class="rb-c-user-badge"'
            '      role="listitem">'
            ' My Badge'
            '</span>')

    def test_render_to_string_with_css_class(self) -> None:
        """Testing UserBadge.render_to_string with css_class"""
        user = self.create_user()
        badge = UserBadge(user=user,
                          label='My Badge',
                          css_class='class1 class2')

        self.assertHTMLEqual(
            badge.render_to_string(),
            '<span class="rb-c-user-badge class1 class2"'
            '      role="listitem">'
            ' My Badge'
            '</span>')
